# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2009 Eugen Wintersberger <eugen.wintersberger@desy.de>
# Copyright (C) 2009-2020 Dominik Kriegner <dominik.kriegner@gmail.com>

"""
module to handle the access to the optical parameters database
"""

import lzma
import os.path
import re

import h5py
import numpy
import scipy.constants


class DataBase(object):

    def __init__(self, fname):
        self.fname = fname
        self.h5file = None  # HDF5 file object holding the database
        self.h5group = None  # Group pointing to the actual element
        self.f0_params = None
        self.f1_en = None
        self.f1 = None
        self.f2_en = None
        self.f2 = None
        self.weight = None
        self.color = None
        self.radius = numpy.nan
        self.matname = None

    def Create(self, dbname, dbdesc):
        """
        Creates a new database. If the database file already exists
        its content is delete.

        Parameters
        ----------
        dbname :    str
            name of the database
        dbdesc :    str
            a short description of the database
        """
        if self.h5file is not None:
            print("database already opened - "
                  "close first to create new database")
            return None

        # tryp to open the database file
        try:
            self.h5file = h5py.File(self.fname, 'w')
        except OSError:
            print('cannot create database file %s!' % (self.fname))
            raise

        # set attributes to the root group with database name and
        # description
        self.h5file.attrs['DBName'] = dbname
        self.h5file.attrs['DBDesc'] = dbdesc

    def Open(self, mode='r'):
        """
        Open an existing database file.
        """
        if self.h5file is not None:
            print('database already opened - '
                  'close first to open new database!')
            return

        try:
            self.h5file = h5py.File(self.fname, mode)
        except OSError:
            print("cannot open database file %s!" % (self.fname))

    def Close(self):
        """
        Close an opend database file.
        """
        if self.h5file is None:
            print("no database file opened!")
            return

        self.h5file.close()
        self.h5file = None

    def CreateMaterial(self, name, description):
        """
        This method creates a new material. If the material group already
        exists the procedure is aborted.

        Parameters
        ----------
        name :          str
            name of the material
        description :   str
            description of the material
        """
        if self.h5file is None:
            print("no database file opened!")
            return

        if name in self.h5file:
            # if the material node already exists a warning message is printed
            print("material node already exists")
        else:
            g = self.h5file.create_group(name)
            g.attrs['name'] = description

    def SetWeight(self, weight):
        """
        Save weight of the element as float

        Parameters
        ----------
        weight :    float
            atomic standard weight of the element
        """
        if not isinstance(weight, float):
            raise TypeError("weight parameter must be a float!")

        self.h5group.attrs['atomic_standard_weight'] = weight
        self.h5file.flush()

    def SetColor(self, color):
        """
        Save color of the element for visualization

        Parameters
        ----------
        color :    tuple, str
            matplotlib color for the element
        """
        if not isinstance(color, (tuple, str)):
            raise TypeError("color parameter must be a tuple or str!")

        self.h5group.attrs['color'] = color
        self.h5file.flush()

    def SetRadius(self, radius):
        """
        Save atomic radius for visualization

        Parameters
        ----------
        radius:     float
            atomic radius in angstrom
        """
        if not isinstance(radius, (float, int)):
            raise TypeError("radius parameter must be a number!")

        self.h5group.attrs['atomic_radius'] = float(radius)
        self.h5file.flush()

    def SetF0(self, parameters, subset='default'):
        """
        Save f0 fit parameters for the set material. The fit parameters
        are stored in the following order:
        c, a1, b1,......., a4, b4

        Parameters
        ----------
        parameters :    list or array-like
            fit parameters
        subset :        str, optional
            name the f0 dataset
        """
        if isinstance(parameters, list):
            p = numpy.array(parameters, dtype=numpy.float32)
        elif isinstance(parameters, numpy.ndarray):
            p = parameters.astype(numpy.float32)
        else:
            raise TypeError("f0 fit parameters must be a "
                            "list or a numpy array!")

        if not subset:
            subset = 'default'

        try:
            del self.h5group['f0/%s' % subset]
        except KeyError:
            pass

        self.h5group.create_dataset('f0/%s' % subset, data=p)
        self.h5file.flush()

    def SetF1F2(self, en, f1, f2):
        """
        Set f1, f2 values for the active material.

        Parameters
        ----------
        en :    list or array-like
            energy in (eV)
        f1 :    list or array-like
            f1 values
        f2 :    list or array-like
            f2 values
        """
        if isinstance(en, (list, tuple)):
            end = numpy.array(en, dtype=numpy.float32)
        elif isinstance(en, numpy.ndarray):
            end = en.astype(numpy.float32)
        else:
            raise TypeError("energy values must be a list or a numpy array!")

        if isinstance(f1, (list, tuple)):
            f1d = numpy.array(f1, dtype=numpy.float32)
        elif isinstance(f1, numpy.ndarray):
            f1d = f1.astype(numpy.float32)
        else:
            raise TypeError("f1 values must be a list or a numpy array!")

        if isinstance(f2, (list, tuple)):
            f2d = numpy.array(f2, dtype=numpy.float32)
        elif isinstance(f2, numpy.ndarray):
            f2d = f2.astype(numpy.float32)
        else:
            raise TypeError("f2 values must be a list or a numpy array!")

        try:
            del self.h5group['en_f12']
        except KeyError:
            pass

        try:
            del self.h5group['f1']
        except KeyError:
            pass

        try:
            del self.h5group['f2']
        except KeyError:
            pass

        self.h5group.create_dataset('en_f12', data=end)
        self.h5group.create_dataset('f1', data=f1d)
        self.h5group.create_dataset('f2', data=f2d)
        self.h5file.flush()

    def SetMaterial(self, name):
        """
        Set a particular material in the database as the actual material.  All
        operations like setting and getting optical constants are done for this
        particular material.

        Parameters
        ----------
        name :  str
            name of the material
        """
        if self.matname == name:
            return
        try:
            self.h5group = self.h5file[name]
        except KeyError:
            print("XU.materials.database: material '%s' not existing!" % name)

        try:
            self.f0_params = self.h5group['f0']
        except KeyError:
            self.f0_params = None
        try:
            self.f1_en = self.h5group['en_f12']
            self.f1 = self.h5group['f1']
        except KeyError:
            self.f1_en = None
            self.f1 = None
        try:
            self.f2_en = self.h5group['en_f12']
            self.f2 = self.h5group['f2']
        except KeyError:
            self.f2_en = None
            self.f2 = None
        try:
            self.weight = self.h5group.attrs['atomic_standard_weight']
        except KeyError:
            self.weight = None
        try:
            self.radius = self.h5group.attrs['atomic_radius']
        except KeyError:
            self.radius = numpy.nan
        try:
            self.color = self.h5group.attrs['color']
        except KeyError:
            self.color = None
        self.matname = name

    def GetF0(self, q, dset='default'):
        """
        Obtain the f0 scattering factor component for a particular
        momentum transfer q.

        Parameters
        ----------
        q :     float or array-like
            momentum transfer
        dset :  str, optional
            specifies which dataset (different oxidation states)
            should be used
        """
        # get parameters from file
        if not dset:
            dset = 'default'
        f0_params = self.f0_params[dset]
        # calculate f0
        if isinstance(q, (numpy.ndarray, list, tuple)):
            ql = numpy.asarray(q)
            f0 = f0_params[0] * numpy.ones(ql.shape)
        else:
            ql = q
            f0 = f0_params[0]
        k = ql / (4. * numpy.pi)

        for i in range(1, len(f0_params) - 1, 2):
            a = f0_params[i]
            b = f0_params[i + 1]
            f0 += a * numpy.exp(-b * k ** 2)

        return f0

    def GetF1(self, en):
        """
        Return the second, energy dependent, real part of the scattering
        factor for a certain energy en.

        Parameters
        ----------
        en :    float or array-like
            energy
        """
        if1 = numpy.interp(en, self.f1_en, self.f1,
                           left=numpy.nan, right=numpy.nan)

        return if1

    def GetF2(self, en):
        """
        Return the imaginary part of the scattering
        factor for a certain energy en.

        Parameters
        ----------
        en :    float or array-like
            energy
        """
        if2 = numpy.interp(en, self.f2_en, self.f2,
                           left=numpy.nan, right=numpy.nan)

        return if2


def init_material_db(db):
    db.CreateMaterial("Dummy", "Dummy atom")
    db.CreateMaterial("H", "Hydrogen")
    db.CreateMaterial("D", "Deuterium")
    db.CreateMaterial("T", "Tritium")
    db.CreateMaterial("He", "Helium")
    db.CreateMaterial("Li", "Lithium")
    db.CreateMaterial("Be", "Berylium")
    db.CreateMaterial("B", "Bor")
    db.CreateMaterial("C", "Carbon")
    db.CreateMaterial("N", "Nitrogen")
    db.CreateMaterial("O", "Oxygen")
    db.CreateMaterial("F", "Flourine")
    db.CreateMaterial("Ne", "Neon")
    db.CreateMaterial("Na", "Sodium")
    db.CreateMaterial("Mg", "Magnesium")
    db.CreateMaterial("Al", "Aluminium")
    db.CreateMaterial("Si", "Silicon")
    db.CreateMaterial("P", "Phosphorus")
    db.CreateMaterial("S", "Sulfur")
    db.CreateMaterial("Cl", "Chlorine")
    db.CreateMaterial("Ar", "Argon")
    db.CreateMaterial("K", "Potassium")
    db.CreateMaterial("Ca", "Calcium")
    db.CreateMaterial("Sc", "Scandium")
    db.CreateMaterial("Ti", "Titanium")
    db.CreateMaterial("V", "Vanadium")
    db.CreateMaterial("Cr", "Chromium")
    db.CreateMaterial("Mn", "Manganese")
    db.CreateMaterial("Fe", "Iron")
    db.CreateMaterial("Co", "Cobalt")
    db.CreateMaterial("Ni", "Nickel")
    db.CreateMaterial("Cu", "Copper")
    db.CreateMaterial("Zn", "Zinc")
    db.CreateMaterial("Ga", "Gallium")
    db.CreateMaterial("Ge", "Germanium")
    db.CreateMaterial("As", "Arsenic")
    db.CreateMaterial("Se", "Selenium")
    db.CreateMaterial("Br", "Bromine")
    db.CreateMaterial("Kr", "Krypton")
    db.CreateMaterial("Rb", "Rubidium")
    db.CreateMaterial("Sr", "Strontium")
    db.CreateMaterial("Y", "Yttrium")
    db.CreateMaterial("Zr", "Zirconium")
    db.CreateMaterial("Nb", "Niobium")
    db.CreateMaterial("Mo", "Molybdenum")
    db.CreateMaterial("Tc", "Technetium")
    db.CreateMaterial("Ru", "Ruthenium")
    db.CreateMaterial("Rh", "Rhodium")
    db.CreateMaterial("Pd", "Palladium")
    db.CreateMaterial("Ag", "Silver")
    db.CreateMaterial("Cd", "Cadmium")
    db.CreateMaterial("In", "Indium")
    db.CreateMaterial("Sn", "Tin")
    db.CreateMaterial("Sb", "Antimony")
    db.CreateMaterial("Te", "Tellurium")
    db.CreateMaterial("I", "Iodine")
    db.CreateMaterial("Xe", "Xenon")
    db.CreateMaterial("Cs", "Caesium")
    db.CreateMaterial("Ba", "Barium")
    db.CreateMaterial("La", "Lanthanum")
    db.CreateMaterial("Ce", "Cerium")
    db.CreateMaterial("Pr", "Praseordymium")
    db.CreateMaterial("Nd", "Neodymium")
    db.CreateMaterial("Pm", "Promethium")
    db.CreateMaterial("Sm", "Samarium")
    db.CreateMaterial("Eu", "Europium")
    db.CreateMaterial("Gd", "Gadolinium")
    db.CreateMaterial("Tb", "Terbium")
    db.CreateMaterial("Dy", "Dysprosium")
    db.CreateMaterial("Ho", "Holmium")
    db.CreateMaterial("Er", "Erbium")
    db.CreateMaterial("Tm", "Thulium")
    db.CreateMaterial("Yb", "Ytterbium")
    db.CreateMaterial("Lu", "Lutetium")
    db.CreateMaterial("Hf", "Hafnium")
    db.CreateMaterial("Ta", "Tantalum")
    db.CreateMaterial("W", "Tungsten")
    db.CreateMaterial("Re", "Rhenium")
    db.CreateMaterial("Os", "Osmium")
    db.CreateMaterial("Ir", "Iridium")
    db.CreateMaterial("Pt", "Platinum")
    db.CreateMaterial("Au", "Gold")
    db.CreateMaterial("Hg", "Mercury")
    db.CreateMaterial("Tl", "Thallium")
    db.CreateMaterial("Pb", "Lead")
    db.CreateMaterial("Bi", "Bismuth")
    db.CreateMaterial("Po", "Polonium")
    db.CreateMaterial("At", "Astatine")
    db.CreateMaterial("Rn", "Radon")
    db.CreateMaterial("Fr", "Fancium")
    db.CreateMaterial("Ra", "Radium")
    db.CreateMaterial("Ac", "Actinium")
    db.CreateMaterial("Th", "Thorium")
    db.CreateMaterial("Pa", "Protactinium")
    db.CreateMaterial("U", "Urianium")
    db.CreateMaterial("Np", "Neptunium")
    db.CreateMaterial("Pu", "Plutonium")
    db.CreateMaterial("Am", "Americium")
    db.CreateMaterial("Cm", "Curium")
    db.CreateMaterial("Bk", "Berkelium")
    db.CreateMaterial("Cf", "Californium")
    db.CreateMaterial("Es", "Einsteinium")
    db.CreateMaterial("Fm", "Fermium")
    db.CreateMaterial("Md", "Mendelevium")
    db.CreateMaterial("No", "Nobelium")
    db.CreateMaterial("Lr", "Lawrencium")
    db.CreateMaterial("Rf", "Rutherfordium")
    db.CreateMaterial("Db", "Dubnium")
    db.CreateMaterial("Sg", "Seaborgium")
    db.CreateMaterial("Bh", "Bohrium")
    db.CreateMaterial("Hs", "Hassium")
    db.CreateMaterial("Mt", "Meitnerium")
    db.CreateMaterial("Ds", "Darmstadtium")
    db.CreateMaterial("Rg", "Roentgenium")
    db.CreateMaterial("Cn", "Copernicium")
    db.CreateMaterial("Nh", "Nihonium")
    db.CreateMaterial("Fl", "Flerovium")
    db.CreateMaterial("Mc", "Moscovium")
    db.CreateMaterial("Lv", "Livermorium")
    db.CreateMaterial("Ts", "Tennessine")
    db.CreateMaterial("Og", "Oganesson")


# functions to read database files
def add_f0_from_intertab(db, itf, verbose=False):
    """
    Read f0 data from International Tables of Crystallography and add
    it to the database.
    """
    # some regular expressions
    elementstr = re.compile(r"^#S")
    multiblank = re.compile(r"\s+")
    while True:
        lb = itf.readline().decode("utf-8")
        if lb == "":
            break
        lb = lb.strip()

        if elementstr.match(lb):
            # found new element
            lb = multiblank.split(lb)

            # determine oxidation state and element name
            elemstate = re.sub('[A-Za-z]', '', lb[2])
            for r, o in zip(('dot', 'p', 'm'), ('.', '+', '-')):
                elemstate = elemstate.replace(o, r)
            if elemstate == 'p2':  # fix wrong name in the source file
                elemstate = '2p'
            ename = re.sub('[^A-Za-z]', '', lb[2])

            if verbose:
                print("{pyname} = Atom('{name}', {num})".format(
                    pyname=ename+elemstate, name=lb[2], num=lb[1]))
            db.SetMaterial(ename)
            # make two dummy reads
            for i in range(2):
                itf.readline()
            # read fit parameters
            lb = itf.readline().decode("utf-8")
            lb = lb.strip()
            lb = multiblank.split(lb)
            a1 = float(lb[0])
            a2 = float(lb[1])
            a3 = float(lb[2])
            a4 = float(lb[3])
            c = float(lb[4])
            b1 = float(lb[5])
            b2 = float(lb[6])
            b3 = float(lb[7])
            b4 = float(lb[8])
            db.SetF0([c, a1, b1, a2, b2, a3, b3, a4, b4], subset=elemstate)


def add_f0_from_xop(db, xop, verbose=False):
    """
    Read f0 data from f0_xop.dat and add
    it to the database.
    """
    # some regular expressions
    elementstr = re.compile(r"^#S")
    multiblank = re.compile(r"\s+")

    while True:
        lb = xop.readline().decode("utf-8")
        if lb == "":
            break
        lb = lb.strip()

        if elementstr.match(lb):
            # found new element
            lb = multiblank.split(lb)
            # determine oxidation state and element name
            elemstate = re.sub('[A-Za-z]', '', lb[2])
            for r, o in zip(('dot', 'p', 'm'), ('.', '+', '-')):
                elemstate = elemstate.replace(o, r)
            ename = re.sub('[^A-Za-z]', '', lb[2])

            if verbose:
                print("{pyname} = Atom('{name}', {num})".format(
                    pyname=ename+elemstate, name=lb[2], num=lb[1]))
            db.SetMaterial(ename)

            # make nine dummy reads
            for i in range(9):
                xop.readline()
            # read fit parameters
            lb = xop.readline().decode("utf-8")
            lb = lb.strip()
            lb = multiblank.split(lb)
            a1 = float(lb[0])
            a2 = float(lb[1])
            a3 = float(lb[2])
            a4 = float(lb[3])
            a5 = float(lb[4])
            c = float(lb[5])
            b1 = float(lb[6])
            b2 = float(lb[7])
            b3 = float(lb[8])
            b4 = float(lb[9])
            b5 = float(lb[10])
            db.SetF0([c, a1, b1, a2, b2, a3, b3, a4, b4, a5, b5])


def add_f1f2_from_henkedb(db, hf, verbose=False):
    """
    Read f1 and f2 data from Henke database and add
    it to the database.
    """
    # some regular expressions
    elementstr = re.compile(r"^#S")
    multiblank = re.compile(r"\s+")
    invalidelem = re.compile(r"[^A-Za-z]")

    while True:
        lb = hf.readline().decode("utf-8")
        if lb == "":
            break
        lb = lb.strip()

        if elementstr.match(lb):
            # found new element
            lb = multiblank.split(lb)
            enum = lb[1]
            ename = lb[2]
            # check if this is not some funny isotope

            if invalidelem.findall(ename) == []:
                if verbose:
                    print("set element %s" % ename)
                db.SetMaterial(ename)
                # make one dummy read
                for i in range(5):
                    hf.readline()

                # read data
                en_list = []
                f1_list = []
                f2_list = []
                while True:
                    lb = hf.readline().decode("utf-8")
                    lb = lb.strip()
                    lb = multiblank.split(lb)
                    en = float(lb[0])
                    # to account for wrong f1 definition in Henke db
                    f1 = float(lb[1]) - float(enum)
                    f2 = float(lb[2])
                    en_list.append(en)
                    f1_list.append(f1)
                    f2_list.append(f2)
                    if en == 30000.:
                        db.SetF1F2(en_list, f1_list, f2_list)
                        break


def add_f1f2_from_kissel(db, kf, verbose=False):
    """
    Read f1 and f2 data from Henke database and add
    it to the database.
    """
    # some regular expressions
    elementstr = re.compile(r"^#S")
    multiblank = re.compile(r"\s+")
    invalidelem = re.compile(r"[^A-Za-z]")

    while True:
        lb = kf.readline().decode("utf-8")
        if lb == "":
            break
        lb = lb.strip()

        if elementstr.match(lb):
            # found new element
            lb = multiblank.split(lb)
            enum = lb[1]
            ename = lb[2]
            # check if this is not some funny isotope

            if invalidelem.findall(ename) == []:
                if verbose:
                    print("set element %s" % ename)
                db.SetMaterial(ename)
                # make 28 dummy reads
                for i in range(28):
                    kf.readline()

                # read data
                en_list = []
                f1_list = []
                f2_list = []
                while True:
                    lb = kf.readline().decode("utf-8")
                    lb = lb.strip()
                    lb = multiblank.split(lb)
                    en = float(lb[0]) * 1000  # convert energy
                    # to account for wrong f1 definition in Henke db
                    f1 = float(lb[4]) - float(enum)
                    f2 = float(lb[5])
                    en_list.append(en)
                    f1_list.append(f1)
                    f2_list.append(f2)
                    if en == 10000000.:
                        db.SetF1F2(en_list, f1_list, f2_list)
                        break


def add_f1f2_from_ascii_file(db, asciifile, element, verbose=False):
    """
    Read f1 and f2 data for specific element from ASCII file (3 columns) and
    save it to the database.
    """

    # parse the f1f2 file
    try:
        af = numpy.loadtxt(asciifile)
    except OSError:
        print("cannot open f1f2 database file")
        return None
    db.SetMaterial(element)

    en = af[:, 0]
    f1 = af[:, 1]
    f2 = af[:, 2]
    db.SetF1F2(en, f1, f2)


def add_mass_from_NIST(db, nistfile, verbose=False):
    """
    Read atoms standard mass and save it to the database.
    The mass of the natural isotope mixture is taken from the NIST data!
    """
    # some regular expressions
    isotope = re.compile(r"^Atomic Number =")
    standardw = re.compile(r"^Standard Atomic Weight")
    relativew = re.compile(r"^Relative Atomic Mass")
    number = re.compile(r"[0-9.]+")
    multiblank = re.compile(r"\s+")

    # parse the nist file
    with open(nistfile, "r") as nf:
        while True:
            lb = nf.readline()
            if lb == "":
                break
            lb = lb.strip()

            if isotope.match(lb):
                # found new element
                lb = multiblank.split(lb)
                lb = nf.readline()
                lb = lb.strip()
                lb = multiblank.split(lb)
                ename = lb[-1]

                if verbose:
                    print("set element %s" % ename)
                db.SetMaterial(ename)

                # read data
                while True:
                    lb = nf.readline()
                    lb = lb.strip()
                    if relativew.match(lb):
                        lb = multiblank.split(lb)
                        # extract fallback weight
                        w = float(number.findall(lb[-1])[0])
                        db.SetWeight(w * scipy.constants.atomic_mass)
                    elif standardw.match(lb):
                        lb = multiblank.split(lb)
                        # extract average weight
                        try:
                            w = float(number.findall(lb[-1])[0])
                            db.SetWeight(w * scipy.constants.atomic_mass)
                        except IndexError:
                            pass
                        break


def add_color_from_JMOL(db, cfile, verbose=False):
    """
    Read color from JMOL color table and save it to the database.
    """
    with open(cfile, "r") as f:
        for line in f.readlines():
            s = line.split()
            ename = s[1]
            color = [float(num)/255. for num in s[2].strip('[]').split(',')]
            color = tuple(color)
            if verbose:
                print("set element %s" % ename)
            db.SetMaterial(ename)
            db.SetColor(color)


def add_radius_from_WIKI(db, dfile, verbose=False):
    """
    Read radius from Wikipedia radius table and save it to the database.
    """
    with open(dfile, "r") as f:
        for line in f.readlines():
            s = line.split(',')
            ename = s[1]
            radius = float(s[3]) / 100.
            if verbose:
                print("set element %s" % ename)
            db.SetMaterial(ename)
            db.SetRadius(radius)


def createAndFillDatabase(fname, dpath=None, verbose=False):
    """
    function to create the database and fill it with values from the various
    source files.

    Parameters
    ----------
    fname : str
        Filename of the database to be created (including the path)
    dpath : str, optional
        directory where all the source data files are stored
    verbose : bool, optional
        flag to determine the verbosity of the script (default: False)
    """
    if dpath is None:
        dpath = os.path.join(os.path.dirname(__file__), 'data')

    dbf = DataBase(fname)
    dbf.Create('elementdata',
               'Database with elemental data from XOP and Kissel databases')

    init_material_db(dbf)

    # add a dummy element, this is useful not only for testing and should be
    # kept in future! It can be used for structure factor calculation tests,
    # and shows how the a database entry can be generated manually
    dbf.SetMaterial('Dummy')
    dbf.SetF0([1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])  # atomic structure factors
    dbf.SetF1F2((0, 1e5), (0, 0), (0, 0))  # zero dispersion correction
    dbf.SetWeight(scipy.constants.atomic_mass)
    dbf.SetRadius(1)

    add_mass_from_NIST(dbf, os.path.join(dpath, 'nist_atom.dat'), verbose)
    add_color_from_JMOL(dbf, os.path.join(dpath, 'colors.dat'), verbose)
    add_radius_from_WIKI(dbf, os.path.join(dpath, 'atomic_radius.dat'),
                         verbose)

    # add F0(Q) for every element
    # with lzma.open(os.path.join('data', 'f0_xop.dat.xz'), 'r') as xop:
    #    add_f0_from_xop(dbf, xop, verbose)
    with lzma.open(os.path.join(dpath, 'f0_InterTables.dat.xz'), 'r') as itf:
        add_f0_from_intertab(dbf, itf, verbose)

    # add F1 and F2 from database
    with lzma.open(os.path.join(dpath, 'f1f2_asf_Kissel.dat.xz'), 'r') as kf:
        add_f1f2_from_kissel(dbf, kf, verbose)
    # with lzma.open(os.path.join(dpath, 'f1f2_Henke.dat'), 'r') as hf:
    #    add_f1f2_from_henkedb(dbf, hf, verbose)

    # Also its possible to add custom data from different databases; e.g.
    # created by Hepaestus (http://bruceravel.github.io/demeter/). This is also
    # possible for specific elements only, therefore extract the data from
    # Hephaestus or any other source producing ASCII files with three columns
    # (energy (eV), f1, f2). To import such data use:
    # add_f1f2_from_ascii_file(dbf, os.path.join(dpath, 'Ga.f1f2'), 'Ga',
    #                          verbose)
    dbf.Close()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2015 Dominik Kriegner <dominik.kriegner@gmail.com>

import unittest

import numpy
import xrayutilities as xu


class TestTransforms(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.x = (1.0, 0.0, 0.0)
        cls.y = (0.0, 1.0, 0.0)
        cls.z = (0.0, 0.0, 1.0)

    def test_Xrot(self):
        xr = xu.math.XRotation(90)
        for i in range(3):
            self.assertAlmostEqual(xr(self.y)[i], self.z[i], places=10)
        for i in range(3):
            self.assertAlmostEqual(xr(self.z)[i], -self.y[i], places=10)

    def test_Yrot(self):
        yr = xu.math.YRotation(90)
        for i in range(3):
            self.assertAlmostEqual(yr(self.z)[i], self.x[i], places=10)
        for i in range(3):
            self.assertAlmostEqual(yr(self.x)[i], -self.z[i], places=10)

    def test_Zrot(self):
        zr = xu.math.ZRotation(90)
        for i in range(3):
            self.assertAlmostEqual(zr(self.x)[i], self.y[i], places=10)
        for i in range(3):
            self.assertAlmostEqual(zr(self.y)[i], -self.x[i], places=10)

    def test_arbrot(self):
        r = xu.math.rotarb
        a = (1, 1, 1)
        for i in range(3):
            self.assertAlmostEqual(r(self.x, a, 120)[i], self.y[i], places=10)
        for i in range(3):
            self.assertAlmostEqual(r(self.z, a, 120)[i], self.x[i], places=10)

    def test_Axis2Z(self):
        a = numpy.random.rand(3)
        a[2] = 2  # ensure non-zero vector
        r = xu.math.AxisToZ(a)
        for i in range(3):
            self.assertAlmostEqual(r(a)[i],
                                   self.z[i] * numpy.linalg.norm(a),
                                   places=10)
        r = xu.math.AxisToZ_keepXY(a)
        for i in range(3):
            self.assertAlmostEqual(r(a)[i],
                                   self.z[i] * numpy.linalg.norm(a),
                                   places=10)
        # test inverse
        for i in range(3):
            self.assertAlmostEqual(r.inverse(self.z)[i],
                                   a[i] / numpy.linalg.norm(a),
                                   places=10)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2020 Dominik Kriegner <dominik.kriegner@gmail.com>

import unittest

import xrayutilities as xu


class Test_findsym(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.materials = []
        for name, obj in xu.materials.predefined_materials.__dict__.items():
            if isinstance(obj, xu.materials.Crystal):
                cls.materials.append(obj)

    def test_findsym(self):
        """
        Test that built in materials use the highest possible space group
        setting for their given unit cell.
        """
        for m in self.materials:
            p1 = m.lattice.convert_to_P1()
            self.assertEqual(m.lattice, p1.findsym(),
                             msg=f"{m.name} does not use highest symmetry!")


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016 Dominik Kriegner <dominik.kriegner@gmail.com>

import unittest

import numpy
import xrayutilities as xu

digits = 11


class TestOpticalProperties(unittest.TestCase):
    en = 'CuKa1'
    n = numpy.complex128(1 - 2.3236677077820289e-05 + 3.1809448794498657e-06j)
    cn = 1 - 1.4554622905893488e-05 + 4.3128498279470241e-07j
    rho_cf = 0.5*8900 + 0.5*7874
    mat = xu.materials.Amorphous('CoFe', rho_cf, [('Co', 0.5), ('Fe', 0.5)])
    cmat = xu.materials.GaAs

    def test_idx_refraction(self):
        idx = self.mat.idx_refraction(en=self.en)
        self.assertAlmostEqual(idx, self.n, places=digits)
        idx = self.cmat.idx_refraction(en=self.en)
        self.assertAlmostEqual(idx, self.cn, places=digits)

    def test_delta_beta(self):
        n2 = 1 - self.mat.delta(en=self.en) + 1j * self.mat.ibeta(en=self.en)
        self.assertAlmostEqual(n2, self.n, places=digits)
        n2 = 1 - self.cmat.delta(en=self.en) + 1j * self.cmat.ibeta(en=self.en)
        self.assertAlmostEqual(n2, self.cn, places=digits)

    def test_chi0(self):
        n3 = 1 + self.mat.chi0(en=self.en) / 2.
        self.assertAlmostEqual(n3, numpy.complex128(self.n), places=digits)
        n3 = 1 + self.cmat.chi0(en=self.en) / 2.
        self.assertAlmostEqual(n3, self.cn, places=digits)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016-2017 Dominik Kriegner <dominik.kriegner@gmail.com>

import math
import unittest

import numpy
import xrayutilities as xu


class TestPseudomorphic(unittest.TestCase):
    mA = xu.materials.Si
    mB = xu.materials.SiGe(numpy.random.rand()*0.9 + 0.1)
    relaxation = numpy.random.rand()
    sub = xu.simpack.Layer(mA, numpy.inf)
    lay = xu.simpack.Layer(mB, 100, relaxation=relaxation)

    def test_pseudomoprhic001(self):
        stack = xu.simpack.PseudomorphicStack001('001', self.sub, self.lay)
        mBpseudo = stack[-1].material

        # calc lattice for mBpseudo
        asub = (self.mA.lattice.a + self.mA.lattice.b) / 2.
        abulk = (self.mB.lattice.a + self.mB.lattice.b) / 2.
        apar = asub + (abulk - asub) * self.relaxation
        epar = (apar - abulk) / abulk
        eperp = -2 * self.mB.c12 / self.mB.c11 * epar
        aperp = abulk * (1 + eperp)

        # check that angles are 90deg
        mBpl = mBpseudo.lattice
        for ang in (mBpl.alpha, mBpl.beta, mBpl.gamma):
            self.assertAlmostEqual(ang, 90, places=12)

        # check lattice parameters
        self.assertAlmostEqual(mBpl.a, apar, places=12)
        self.assertAlmostEqual(mBpl.b, apar, places=12)
        self.assertAlmostEqual(mBpl.c, aperp, places=12)

    def test_pseudomorphic111(self):
        stack = xu.simpack.PseudomorphicStack111('111', self.sub + self.lay)
        mBpseudo = stack[-1].material

        # calc lattice for mBpseudo
        def get_inplane111(lay):
            """determine inplane lattice parameter for (111) surfaces"""
            return (xu.math.VecNorm(lay.GetPoint(1, 1, -2)) / math.sqrt(6) +
                    xu.math.VecNorm(lay.GetPoint(1, -1, 0)) / math.sqrt(2)) / 2
        asub = get_inplane111(self.mA.lattice)
        abulk = get_inplane111(self.mB.lattice)

        apar = asub + (abulk - asub) * self.relaxation
        epar = (apar - abulk) / abulk
        eperp = -epar * (2*self.mB.c11 + 4*self.mB.c12 - 4*self.mB.c44) /\
                        (self.mB.c11 + 2*self.mB.c12 + 4*self.mB.c44)
        # eps = (eperp - epar + 3 * numpy.identity(3) * epar) / 3.

        # check that angles lattice spacings are correct
        pi = numpy.pi
        self.assertAlmostEqual(
            mBpseudo.planeDistance(1, 1, -2),
            2*pi/(xu.math.VecNorm(self.mB.Q(1, 1, -2))/(1+epar)), places=12)
        self.assertAlmostEqual(
            mBpseudo.planeDistance(1, 1, 1),
            2*pi/(xu.math.VecNorm(self.mB.Q(1, 1, 1))/(1+eperp)), places=12)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2015 Dominik Kriegner <dominik.kriegner@gmail.com>

import unittest

import numpy
import xrayutilities as xu


class TestQ2AngFit(unittest.TestCase):
    energy = 15000

    @classmethod
    def setUpClass(cls):
        cls.qconv = xu.experiment.QConversion(['z+', 'y-', 'z-'],
                                              ['z+', 'y-'],
                                              [1, 0, 0])
        cls.hxrd = xu.HXRD((1, 0, 0), (0, 0, 1), en=cls.energy,
                           qconv=cls.qconv)
        cls.bounds = (0, (-180, 180), 0, (-1, 90), (-1, 90))
        qz = numpy.random.rand() + 1
        cls.qvec = numpy.array(((numpy.random.rand()-0.5)*qz, 0, qz))

    def test_q2angfit(self):
        ang, qerror, errcode = xu.Q2AngFit(self.qvec, self.hxrd, self.bounds)
        qout = self.hxrd.Ang2Q(*ang)
        for i in range(3):
            self.assertAlmostEqual(qout[i], self.qvec[i], places=5)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016 Dominik Kriegner <dominik.kriegner@gmail.com>

import os.path
import unittest

import numpy
import xrayutilities as xu

xu.config.VERBOSITY = 0  # make no output during test
datadir = os.path.join(os.path.dirname(__file__), 'data')
fullfilename = os.path.join(datadir, "detalign.xrdml.bz2")


@unittest.skipIf(not os.path.isfile(fullfilename),
                 "additional test data needed (http://xrayutilities.sf.io)")
class TestLinear_calib(unittest.TestCase):
    pw0 = 1.4813e-04
    cch0 = 633.90
    tilt0 = -0.498

    @classmethod
    def setUpClass(cls):
        tt, det = xu.io.getxrdml_scan(fullfilename)
        cls.ang = tt[:, 639]
        cls.spectra = det

    def test_linear_calib(self):
        pwidth, cch, tilt = xu.analysis.linear_detector_calib(self.ang,
                                                              self.spectra)
        self.assertAlmostEqual(numpy.abs(pwidth), self.pw0, places=7)
        self.assertAlmostEqual(cch, self.cch0, places=1)
        self.assertAlmostEqual(tilt, self.tilt0, places=2)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2010-2011 Dominik Kriegner <dominik.kriegner@gmail.com>

"""
module to provide functions that perform block averaging
of intensity arrays to reduce the amount of data (mainly
for PSD and CCD measurements

and

provide functions for normalizing intensities for

* count time
* absorber (user-defined function)
* monitor
* flatfield correction
"""

import numpy

from . import config, cxrayutilities, math, utilities
from .exception import InputError


def blockAverage1D(data, Nav):
    """
    perform block average for 1D array/list of Scalar values
    all data are used. at the end of the array a smaller cell
    may be used by the averaging algorithm

    Parameters
    ----------
    data :  array-like
        data which should be contracted (length N)
    Nav :   int
        number of values which should be averaged

    Returns
    -------
    ndarray
        block averaged numpy array of data type numpy.double
        (length ceil(N/Nav))
    """

    if not isinstance(data, (numpy.ndarray, list)):
        raise TypeError("first argument data must be of type list or "
                        "numpy.ndarray")

    data = numpy.array(data, dtype=numpy.double)
    block_av = cxrayutilities.block_average1d(data, Nav)

    return block_av


def blockAverage2D(data2d, Nav1, Nav2, **kwargs):
    """
    perform a block average for 2D array of Scalar values
    all data are used therefore the margin cells may differ in size

    Parameters
    ----------
    data2d :        ndarray
        array of 2D data shape (N, M)
    Nav1, Nav2 :    int
        a field of (Nav1 x Nav2) values is contracted

    kwargs :        dict, optional
        optional keyword argument
    roi :           tuple or list, optional
        region of interest for the 2D array. e.g. [20, 980, 40, 960],
        reduces M, and M!

    Returns
    -------
    ndarray
        block averaged numpy array with type numpy.double with shape
        (ceil(N/Nav1), ceil(M/Nav2))
    """

    if not isinstance(data2d, (numpy.ndarray)):
        raise TypeError("first argument data2d must be of type numpy.ndarray")

    roi = kwargs.get('roi', [0, data2d.shape[0], 0, data2d.shape[1]])
    data = numpy.array(data2d[roi[0]:roi[1], roi[2]:roi[3]],
                       dtype=numpy.double)

    if config.VERBOSITY >= config.DEBUG:
        N, M = (roi[1] - roi[0], roi[3] - roi[2])
        print("xu.normalize.blockAverage2D: roi: %s" % (str(roi)))
        print("xu.normalize.blockAverage2D: Nav1, 2: %d,%d" % (Nav1, Nav2))
        print("xu.normalize.blockAverage2D: number of points: (%d,%d)"
              % (numpy.ceil(N / float(Nav1)), numpy.ceil(M / float(Nav2))))

    block_av = cxrayutilities.block_average2d(data, Nav1, Nav2,
                                              config.NTHREADS)

    return block_av


def blockAveragePSD(psddata, Nav, **kwargs):
    """
    perform a block average for serveral PSD spectra
    all data are used therefore the last cell used for
    averaging may differ in size

    Parameters
    ----------
    psddata :   ndarray
        array of 2D data shape (Nspectra, Nchannels)
    Nav :       int
        number of channels which should be averaged

    kwargs :    dict, optional
        optional keyword argument
    roi :       tuple or list
        region of interest for the 2D array. e.g. [20, 980] Nchannels = 980-20

    Returns
    -------
    ndarray
        block averaged psd spectra as numpy array with type numpy.double of
        shape (Nspectra , ceil(Nchannels/Nav))
    """

    if not isinstance(psddata, (numpy.ndarray)):
        raise TypeError("first argument psddata must be of type numpy.ndarray")

    roi = kwargs.get('roi', [0, psddata.shape[1]])
    data = numpy.array(psddata[:, roi[0]:roi[1]], dtype=numpy.double)

    block_av = cxrayutilities.block_average_PSD(data, Nav, config.NTHREADS)

    return block_av


def blockAverageCCD(data3d, Nav1, Nav2, **kwargs):
    """
    perform a block average for 2D frames inside a 3D array.
    all data are used therefore the margin cells may differ in size

    Parameters
    ----------
    data3d :        ndarray
        array of 3D data shape (Nframes, N, M)
    Nav1, Nav2 :    int
        a field of (Nav1 x Nav2) values is contracted

    kwargs :        dict, optional
        optional keyword argument
    roi :           tuple or list, optional
        region of interest for the 2D array. e.g. [20, 980, 40, 960],
        reduces M, and M!

    Returns
    -------
    ndarray
        block averaged numpy array with type numpy.double with shape
        (Nframes, ceil(N/Nav1), ceil(M/Nav2))
    """

    if not isinstance(data3d, (numpy.ndarray)):
        raise TypeError("first argument data3d must be of type numpy.ndarray")

    roi = kwargs.get('roi', [0, data3d.shape[1], 0, data3d.shape[2]])
    data = numpy.array(data3d[:, roi[0]:roi[1], roi[2]:roi[3]],
                       dtype=numpy.double)

    if config.VERBOSITY >= config.DEBUG:
        N, M = (roi[1] - roi[0], roi[3] - roi[2])
        print("xu.normalize.blockAverageCCD: roi: %s" % (str(roi)))
        print("xu.normalize.blockAverageCCD: Nav1, 2: %d,%d" % (Nav1, Nav2))
        print("xu.normalize.blockAverageCCD: number of points: (%d,%d)"
              % (numpy.ceil(N / float(Nav1)), numpy.ceil(M / float(Nav2))))

    block_av = cxrayutilities.block_average_CCD(data, Nav1, Nav2,
                                                config.NTHREADS)

    return block_av

# #####################################
# #    Intensity correction class    ##
# #####################################


class IntensityNormalizer(object):

    """
    generic class for correction of intensity (point detector, or MCA,
    single CCD frames) for count time and absorber factors
    the class must be supplied with a absorber correction function
    and works with data structures provided by xrayutilities.io classes or the
    corresponding objects from hdf5 files
    """

    def __init__(self, det='', **keyargs):
        """
        initialization of the corrector class

        Parameters
        ----------
        det :       str
            detector field name of the data structure

        mon :       str, optional
            monitor field name
        time:       float or str, optional
            count time field name or count time as float
        av_mon :    float, optional
            average monitor value (default: data[mon].mean())
        smoothmon : int
            number of monitor values used to get a smooth monitor signal
        absfun :    callable, optional
            absorber correction function to be used as in
            ``absorber_corrected_intensity = data[det]*absfun(data)``
        flatfield : ndarray
            flatfield of the detector; shape must be the same as data[det], and
            is only applied for MCA detectors
        darkfield : ndarray
            darkfield of the detector; shape must be the same as data[det], and
            is only applied for MCA detectors

        Examples
        --------
        >>> detcorr = IntensityNormalizer("MCA", time="Seconds",
        >>>     absfun=lambda d: d["PSDCORR"]/d["PSD"].astype(numpy.float))
        """
        valid_kwargs = {'mon': 'monitor field name',
                        'time': 'count time field/value',
                        'smoothmon': 'number of monitor values to average',
                        'av_mon': 'average monitor value',
                        'absfun': 'absorber correction function',
                        'flatfield': 'detector flatfield',
                        'darkfield': 'detector darkfield'}
        utilities.check_kwargs(keyargs, valid_kwargs,
                               self.__class__.__name__)

        # check input arguments
        self._setdet(det)

        self._setmon(keyargs.get('mon', None))
        self._settime(keyargs.get('time', None))
        self._setavmon(keyargs.get('av_mon', None))
        self._setabsfun(keyargs.get('absfun', None))
        self._setflatfield(keyargs.get('flatfield', None))
        self._setdarkfield(keyargs.get('darkfield', None))
        self.smoothmon = keyargs.get('smoothmon', 1)

    def _getdet(self):
        """
        det property handler

        returns the detector field name
        """
        return self._det

    def _setdet(self, det):
        """
        det  property handler

        sets the detector field name
        """
        if isinstance(det, str):
            self._det = det
        else:
            self._det = None
            raise TypeError("argument det must be of type str")

    def _gettime(self):
        """
        time property handler

        returns the count time or the field name of the count time
        or None if time is not set
        """
        return self._time

    def _settime(self, time):
        """
        time property handler

        sets the count time field or value
        """
        if isinstance(time, str):
            self._time = time
        elif isinstance(time, (float, int)):
            self._time = float(time)
        elif isinstance(time, type(None)):
            self._time = None
        else:
            raise TypeError("argument time must be of type str, float or None")

    def _getmon(self):
        """
        mon property handler

        returns the monitor field name or None if not set
        """
        return self._mon

    def _setmon(self, mon):
        """
        mon property handler

        sets the monitor field name
        """
        if isinstance(mon, str):
            self._mon = mon
        elif isinstance(mon, type(None)):
            self._mon = None
        else:
            raise TypeError("argument mon must be of type str")

    def _getavmon(self):
        """
        av_mon property handler

        returns the value of the average monitor or None
        if average is calculated from the monitor field
        """
        return self._avmon

    def _setavmon(self, avmon):
        """
        avmon property handler

        sets the average monitor field name
        """
        if isinstance(avmon, (float, int)):
            self._avmon = float(avmon)
        elif isinstance(avmon, type(None)):
            self._avmon = None
        else:
            raise TypeError("argument avmon must be of type float or None")

    def _getabsfun(self):
        """
        absfun property handler

        returns the costum correction function or None
        """
        return self._absfun

    def _setabsfun(self, absfun):
        """
        absfun property handler

        sets the costum correction function
        """
        if hasattr(absfun, '__call__'):
            self._absfun = absfun
            if self._absfun.__code__.co_argcount != 1:
                raise TypeError("argument absfun must be a function with one "
                                "argument (data object)")
        elif isinstance(absfun, type(None)):
            self._absfun = None
        else:
            raise TypeError("argument absfun must be of type function or None")

    def _getflatfield(self):
        """
        flatfield property handler

        returns the current set flatfield of the detector
        or None if not set
        """
        return self._flatfield

    def _setflatfield(self, flatf):
        """
        flatfield property handler

        sets the flatfield of the detector
        """
        if isinstance(flatf, (list, tuple, numpy.ndarray)):
            self._flatfield = numpy.array(flatf, dtype=numpy.float)
            self._flatfieldav = numpy.mean(self._flatfield[
                                           self._flatfield.nonzero()])
            self._flatfield[self.flatfield < 1.e-5] = 1.0
        elif isinstance(flatf, type(None)):
            self._flatfield = None
        else:
            raise TypeError("argument flatfield must be of type list, tuple, "
                            "numpy.ndarray or None")

    def _getdarkfield(self):
        """
        flatfield property handler

        returns the current set darkfield of the detector
        or None if not set
        """
        return self._darkfield

    def _setdarkfield(self, darkf):
        """
        flatfield property handler

        sets the darkfield of the detector
        """
        if isinstance(darkf, (list, tuple, numpy.ndarray)):
            self._darkfield = numpy.array(darkf, dtype=numpy.float)
            self._darkfieldav = numpy.mean(self._darkfield)
        elif isinstance(darkf, type(None)):
            self._darkfield = None
        else:
            raise TypeError("argument flatfield must be of type list, tuple, "
                            "numpy.ndarray or None")

    det = property(_getdet, _setdet)
    time = property(_gettime, _settime)
    mon = property(_getmon, _setmon)
    avmon = property(_getavmon, _setavmon)
    absfun = property(_getabsfun, _setabsfun)
    flatfield = property(_getflatfield, _setflatfield)
    darkfield = property(_getdarkfield, _setdarkfield)

    def __call__(self, data, ccd=None):
        """
        apply the correction method which was initialized to the measured data

        Parameters
        ----------
        data :  numpy.recarray
            data object from xrayutilities.io classes
        ccd :   ndarray, optional
            optionally CCD data can be given as separate ndarray of shape
            (len(data), n1, n2), where n1, n2 is the shape of the CCD image.

        Returns
        -------
        corrint :   ndarray
            corrected intensity as numpy.ndarray of the same shape as data[det]
            (or ccd.shape)
        """
        if numpy.any(ccd):
            rawdata = ccd
        else:
            rawdata = data[self._det]

        corrint = numpy.zeros(rawdata.shape, dtype=numpy.float)

        # set needed variables
        # monitor intensity
        if self._mon:
            if self.smoothmon == 1:
                mon = data[self._mon]
            else:
                mon = math.smooth(data[self._mon], self.smoothmon)
        else:
            mon = 1.
        # count time
        if isinstance(self._time, str):
            time = data[self._time]
        elif isinstance(self._time, float):
            time = self._time
        else:
            time = 1.
        # average monitor counts
        if self._avmon:
            avmon = self._avmon
        else:
            avmon = numpy.mean(mon)
        # absorber correction function
        if self._absfun:
            abscorr = self._absfun(data)
        else:
            abscorr = 1.

        c = abscorr * avmon / (mon * time)
        # correct the correction factor if it was evaluated to an incorrect
        # value
        if isinstance(c, numpy.ndarray):
            c[numpy.isnan(c)] = 1.0
            c[numpy.isinf(c)] = 1.0
            c[c == 0] = 1.0
        else:
            if numpy.isnan(c) or numpy.isinf(c) or c == 0:
                c = 1.0

        if len(rawdata.shape) == 1:
            corrint = rawdata * c
        elif len(rawdata.shape) == 2 and isinstance(c, numpy.ndarray):
            # 1D detector c.shape[0] should be rawdata.shape[0]
            if self._darkfield is not None:
                if self._darkfield.shape[0] != rawdata.shape[1]:
                    raise InputError("data[det] second dimension must have "
                                     "the same length as darkfield")

                if isinstance(time, numpy.ndarray):
                    # darkfield correction
                    corrint = (rawdata -
                               self._darkfield[numpy.newaxis, :] *
                               time[:, numpy.newaxis])
                elif isinstance(time, float):
                    # darkfield correction
                    corrint = (rawdata -
                               self._darkfield[numpy.newaxis, :] * time)
                else:
                    print("XU.normalize.IntensityNormalizer: check "
                          "initialization and your input")
                    return None
                corrint[corrint < 0.] = 0.

            else:
                corrint = rawdata

            corrint = corrint * c[:, numpy.newaxis]

            if self._flatfield is not None:
                if self._flatfield.shape[0] != rawdata.shape[1]:
                    raise InputError("data[det] second dimension must have "
                                     "the same length as flatfield")
                # flatfield correction
                corrint = (corrint / self._flatfield[numpy.newaxis, :] *
                           self._flatfieldav)

        elif len(rawdata.shape) == 2 and isinstance(c, numpy.float):
            # single 2D detector frame
            corrint = rawdata * c

        elif len(rawdata.shape) == 3:
            # darkfield and flatfield correction is still missing!
            corrint = rawdata * c[:, numpy.newaxis, numpy.newaxis]

        else:
            raise InputError("data[det] must be an array of dimension one "
                             "or two or three")

        return corrint



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016-2019 Dominik Kriegner <dominik.kriegner@gmail.com>

import os

import lmfit
import numpy
import xrayutilities as xu
from matplotlib.pylab import *

# load experimental data
ai, edata, eps = numpy.loadtxt(os.path.join('data', 'xrr_data.txt'),
                               unpack=True)
ai /= 2.0

# define layers
# SiO2 / Ru(5) / CoFe(3) / IrMn(3) / AlOx(10)
lSiO2 = xu.simpack.Layer(xu.materials.SiO2, inf, roughness=2.5)
lRu = xu.simpack.Layer(xu.materials.Ru, 47, roughness=2.8)
rho_cf = 0.5*8900 + 0.5*7874
mat_cf = xu.materials.Amorphous('CoFe', rho_cf)
lCoFe = xu.simpack.Layer(mat_cf, 27, roughness=4.6)
lIrMn = xu.simpack.Layer(xu.materials.Ir20Mn80, 21, roughness=3.0)
lAl2O3 = xu.simpack.Layer(xu.materials.Al2O3, 100, roughness=5.5)

# create model
m = xu.simpack.SpecularReflectivityModel(lSiO2, lRu, lCoFe, lIrMn, lAl2O3,
                                         energy='CuKa1', resolution_width=0.02,
                                         sample_width=6, beam_width=0.25,
                                         background=81, I0=6.35e9)

# embed model in fit code
fitm = xu.simpack.FitModel(m, plot=True, verbose=True)

# set some parameter limitations
fitm.set_param_hint('SiO2_density', vary=False)
fitm.set_param_hint('Al2O3_density', min=0.8*xu.materials.Al2O3.density,
                    max=1.2*xu.materials.Al2O3.density)
p = fitm.make_params()
fitm.set_fit_limits(xmin=0.05, xmax=8.0)

# perform the fit
res = fitm.fit(edata, p, ai, weights=1/eps)
lmfit.report_fit(res, min_correl=0.5)

m.densityprofile(500, plot=True)
show()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2012 Dominik Kriegner <dominik.kriegner@gmail.com>

import os.path
from itertools import permutations

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy
import xrayutilities as xu
from mpl_toolkits.basemap import *

datadir = "data"

eps = 0.01
# plot settings
mpl.rcParams['font.size'] = 18.0
mpl.rcParams['axes.labelsize'] = 'large'
mpl.rcParams['legend.fancybox'] = True
mpl.rcParams['legend.handletextpad'] = 0.5
mpl.rcParams['legend.fontsize'] = 'medium'
mpl.rcParams['figure.subplot.bottom'] = 0.13
mpl.rcParams['figure.subplot.top'] = 0.93
mpl.rcParams['figure.subplot.left'] = 0.14
mpl.rcParams['figure.subplot.right'] = 0.915
mpl.rcParams['savefig.dpi'] = 300
mpl.rcParams['axes.grid'] = False

# substrate
Ge = xu.materials.Ge

hkls = list(permutations([1, 1, 3])) + list(permutations([-1, 1, 3])) +\
       list(permutations([-1, -1, 3])) + list(permutations([1, 1, -3])) +\
       list(permutations([-1, 1, -3])) + list(permutations([-1, -1, -3]))

tup = []
label = []
for ind in hkls:
    tup.append(Ge.Q(*ind))
    lab = r'$('
    for index in ind:
        lab += '%s' % (r'\bar' if index < 0 else '') + str(abs(index))
    lab += ')$'
    label.append(lab)

df = xu.io.XRDMLFile(os.path.join(datadir, "polefig_Ge113.xrdml.bz2"))
s = df.scan

chi = 90 - s['Psi']
phi = s['Phi']
INT = s['detector']

# create 2D arrays for the angles
CHI = chi[:, numpy.newaxis] * numpy.ones(INT.shape)
PHI = phi

INT = xu.maplog(INT, 6, 0)

fig = plt.figure()
plt.clf()
# plt.title("(113) pole figure")
m = Basemap(boundinglat=-1., lon_0=180.0, resolution=None,
            projection='npstere')
X, Y = m(PHI, CHI)
ax = plt.subplot(111)
ax.set_frame_on(False)
CS = m.contourf(X, Y, INT, 50)
m.drawparallels(numpy.arange(0, 91, 10), labels=[1, 1, 1, 1],
                color='gray', dashes=[2, 2])
m.drawmeridians(numpy.arange(0, 360, 60), labels=[1, 1, 1, 1],
                color='gray', labelstyle='+/-', dashes=[2, 2])

dphi = 62
dchi = -32
inpdir = xu.math.rotarb(Ge.Q(1, -1, 0), Ge.Q(0, 0, 1), dphi)
ndir = xu.math.rotarb(Ge.Q(0, 0, 1), inpdir, dchi)

# plot normaldir
# calculate spherical coordinate angles
[chi, phi] = xu.analysis.getangles(Ge.Q(0, 0, 1), ndir, inpdir)
chi = 90 - chi
phi = phi - dphi
# if direction is visible plot in polefig
if (chi >= -eps):
    x, y = m(phi, chi)
    m.plot(numpy.array([x]), numpy.array([y]), ls='None', marker='s',
           color='k', ms=12.)

# plot Ge {113} Bragg peaks
for i in range(len(tup)):
    dir = tup[i]
    # calculate spherical coordinate angles
    dir = xu.math.rotarb(dir, Ge.Q(0, 0, 1), 27)
    [chi, phi] = xu.analysis.getangles(dir, ndir, inpdir)
    chi = 90 - chi
    phi = phi - dphi
    # if direction is visible plot in polefig
    if (chi >= -eps):
        x, y = m(phi, chi)
        m.plot(numpy.array([x]), numpy.array([y]), ls='None', marker='o',
               color='k', ms=8., mfc='None', mew=2.)
        plt.text(x + 400000, y, label[i], ha='left', va='bottom')



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2015-2019 Dominik Kriegner <dominik.kriegner@gmail.com>

import os.path
import unittest

import numpy
import xrayutilities as xu

testfile = 'omega_mm.xrdml'
testfile2 = 'cecchi_refl_30min.xrdml'
datadir = os.path.join(os.path.dirname(__file__), 'data')
fullfilename = os.path.join(datadir, testfile)
fullfilename2 = os.path.join(datadir, testfile2)


@unittest.skipIf(not os.path.isfile(fullfilename) or
                 not os.path.isfile(fullfilename2),
                 "additional test data needed (http://xrayutilities.sf.io)")
class TestIO_XRDML(unittest.TestCase):
    dshape = (499,)
    dmax = 75052800.0
    dmin = 54.0
    motmax = 16.7872
    motmin = 11.8072
    tpos = 240
    dtpos = 1502.0

    @classmethod
    def setUpClass(cls):
        cls.xrdmlfile = xu.io.XRDMLFile(fullfilename)
        cls.xrdmlfile2 = xu.io.XRDMLFile(fullfilename2)
        cls.data1 = cls.xrdmlfile.scan['detector']
        cls.motor, _, cls.data2 = xu.io.getxrdml_scan(fullfilename, 'Phi')

    def test_datashape(self):
        self.assertEqual(self.dshape, self.data1.shape)
        self.assertEqual(self.dshape, self.data2.shape)
        self.assertEqual(self.dshape, self.motor.shape)

    def test_datavalues(self):
        self.assertAlmostEqual(self.motmax, self.motor.max(), places=10)
        self.assertAlmostEqual(self.motmin, self.motor.min(), places=10)
        self.assertAlmostEqual(self.dmax, self.data1.max(), places=10)
        self.assertAlmostEqual(self.dmin, self.data1.min(), places=10)
        self.assertAlmostEqual(self.dtpos, self.data1[self.tpos], places=10)

    def test_datamethods(self):
        self.assertTrue(numpy.all(self.data1 == self.data2))

    def test_version2(self):
        self.assertEqual(len(self.xrdmlfile2.scan['counts']),
                         len(self.xrdmlfile2.scan['detector']))


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2010-2020 Dominik Kriegner <dominik.kriegner@gmail.com>

"""
xrayutilities utilities contains a conglomeration of useful functions
which do not fit into one of the other files
"""

import fractions
import math

import numpy

from . import config
from .utilities_noconf import *


def import_matplotlib_pyplot(funcname='XU'):
    """
    lazy import function of matplotlib.pyplot

    Parameters
    ----------
    funcname :      str
        identification string of the calling function

    Returns
    -------
    flag :  bool
        the flag is True if the loading was successful and False otherwise.
    pyplot
        On success pyplot is the matplotlib.pyplot package.
    """
    try:
        from matplotlib import pyplot as plt
        from .mpl_helper import SqrtAllowNegScale
        return True, plt
    except ImportError:
        if config.VERBOSITY >= config.INFO_LOW:
            print("%s: Warning: plot functionality not available" % funcname)
        return False, None


def import_lmfit(funcname='XU'):
    """
    lazy import function for lmfit
    """
    try:
        import lmfit
        return lmfit
    except ImportError:
        raise ImportError("%s: Fitting of models needs the lmfit package "
                          "(https://pypi.python.org/pypi/lmfit)" % funcname)


def import_mayavi_mlab(funcname='XU'):
    """
    lazy import function of mayavi.mlab

    Parameters
    ----------
    funcname :      str
        identification string of the calling function

    Returns
    -------
    flag :  bool
        the flag is True if the loading was successful and False otherwise.
    mlab
        On success mlab is the mayavi.mlab package.
    """
    try:
        from mayavi import mlab
        return True, mlab
    except ImportError:
        if config.VERBOSITY >= config.INFO_LOW:
            print("%s: Warning: plot functionality not available" % funcname)
        return False, None


def maplog(inte, dynlow="config", dynhigh="config"):
    """
    clips values smaller and larger as the given bounds and returns the log10
    of the input array. The bounds are given as exponent with base 10 with
    respect to the maximum in the input array.  The function is implemented in
    analogy to J. Stangl's matlab implementation.

    Parameters
    ----------
    inte :      ndarray
        numpy.array, values to be cut in range
    dynlow :    float, optional
        10^(-dynlow) will be the minimum cut off
    dynhigh :   float, optional
        10^(-dynhigh) will be the maximum cut off

    Returns
    -------
    ndarray
        numpy.array of the same shape as inte, where values smaller/larger than
        10^(-dynlow, dynhigh) were replaced by 10^(-dynlow, dynhigh)

    Examples
    --------
    >>> lint = maplog(int, 5, 2)
    """
    if dynlow == "config":
        dynlow = config.DYNLOW
    if dynhigh == "config":
        dynhigh = config.DYNHIGH

    mask = numpy.logical_not(numpy.isnan(inte))
    if inte[mask].max() <= 0.0:
        raise ValueError("XU.maplog: only negativ or zero values given. "
                         "Log is not defined!")
    ma = inte[mask].max() * 10 ** (-1*dynhigh)  # upper bound
    mi = inte[mask].max() * 10 ** (-1*dynlow)  # lower bound

    return numpy.log10(numpy.minimum(numpy.maximum(inte, mi), ma))


def frac2str(f, denominator_limit=25, fmt='%7.4f'):
    """
    convert a float to a string attempting to represent it as a fraction

    Parameters
    ----------
    f :    float
        floating point number to be represented as string
    denominator_limit : int
        maximal integer used as denominator. If f can't be expressed
        (within xu.config.EPSILON) by a fraction with a denominator
        up to this number a floating point string will be returned
    fmt :  str
        format string used in case a floating point representation is needed

    Returns
    -------
    str
    """
    frac = fractions.Fraction(f).limit_denominator(denominator_limit)
    if math.isclose(float(frac), f, abs_tol=config.EPSILON):
        return str(frac)
    else:
        return fmt % f



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2014 Raphael Grifone <raphael.grifone@esrf.fr>
# Copyright (C) 2014,2018 Dominik Kriegner <dominik.kriegner@gmail.com>

import os

import xrayutilities as xu
from matplotlib.pylab import *

import xrayutilities_id01_functions as id01

# 3S+2D goniometer (ID01 goniometer, sample mu, eta, phi detector
# nu, del, mpxy, mpxz
# convention for coordinate system: x downstream; z upwards; y to the
# "outside" (righthanded)
# QConversion will set up the goniometer geometry. So the first argument
# describes the sample rotations, the second the detector rotations and the
# third the primary beam direction.
# For this consider the following right handed coordinate system (feel free to
# use your conventions):
# x: downstream (direction of primary beam)
# y: out of the ring
# z: upwards
# The outer most sample rotation (so the one mounted on the floor) is one
# which turns left-handed (-) around the z-direction -> z- (mu)
# The second sample rotation ('eta') is lefthanded (-) around y -> y-
qconv = xu.experiment.QConversion(['z-', 'y-', 'z-'],
                                  ['z-', 'y-', 'ty', 'tz'],
                                  [1, 0, 0])
hxrd = xu.HXRD([1, 1, 0], [0, 0, 1], qconv=qconv, sampleor='z+')
hxrd._A2QConversion.init_area('z-', 'y+', cch1=333.94, cch2=235.62, Nch1=516,
                              Nch2=516, pwidth1=5.5000e-02, pwidth2=5.5000e-02,
                              distance=0.53588*1000, detrot=-1.495,
                              tiltazimuth=155.0, tilt=0.745, Nav=(2, 2))

#############################################################
# load spec-file data
fnames = ('KMAP_2017_fast_00001.spec',)

scannrs = []
for fn in fnames:
    sf = xu.io.SPECFile(fn, path=datadir)
    scannrs.append(list(range(len(sf))))

nx, ny = 150, 150

# specfile, scannumbers, nx,ny, motornames, optional column names (ID01
# values are default)
fss = xu.io.FastScanSeries(fnames, scannrs, nx, ny, 'mu', 'eta', 'phi', 'nu',
                           'del', 'mpxy', 'mpxz', xmotor='adcY', ymotor='adcX',
                           ccdnr='imgnr', path=datadir)

#############################################################
# 3D RSM from summing over X,Y
# now all EDF files are parsed, this will take a while and some memory
qconv.energy = id01.getmono_energy(sf[0])
xu.config.VERBOSITY = 2
g3d = fss.get_average_RSM(81, 82, 83, qconv, datadir=id01.datadir,
                          replacedir=id01.repl_n, roi=None, nav=(4, 4),
                          filterfunc=deadpixelkill)
xu.config.VERBOSITY = 1
numpy.savez_compressed('RSM3D.npz', qx=g3d.xaxis, qy=g3d.yaxis, qz=g3d.zaxis,
                       data=g3d.data)


figure()
subplot(221)
pcolormesh(qx, qy, data.sum(axis=2).T, norm=mpl.colors.LogNorm())
xlabel(r'Qx ($\mathrm{\AA}^{-1}$)')
ylabel(r'Qy ($\mathrm{\AA}^{-1}$)')

subplot(222)
pcolormesh(qx, qz, data.sum(axis=1).T, norm=mpl.colors.LogNorm())
xlabel(r'Qx ($\mathrm{\AA}^{-1}$)')
ylabel(r'Qz ($\mathrm{\AA}^{-1}$)')

subplot(223)
pcolormesh(qy, qz, data.sum(axis=0).T, norm=mpl.colors.LogNorm())
xlabel(r'Qy ($\mathrm{\AA}^{-1}$)')
ylabel(r'Qz ($\mathrm{\AA}^{-1}$)')
tight_layout()

#############################################################
# 2D real space maps for selected region in Q-space
qr = [0.57, 0.62, -0.20, -0.16, 3.47, 3.50]  # [xmin, xmax, ymin, ..., zmax]
x, y, data = fss.get_sxrd_for_qrange(qr, qconv, datadir=id01.datadir,
                                     replacedir=id01.repl_n)
numpy.savez_compressed('output_sxrd_map.npz', x=x, y=y, data=data)


figure()
lev_exp = np.linspace(np.log10(data.min()),
                      np.log10(data.max()), 100)
levs = np.power(10, lev_exp)
tricontourf(y, x, data, levs, norm=mpl.colors.LogNorm())
axis('scaled')
ylabel('piy (um)')
xlabel('pix (um)')
tight_layout()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2015 Dominik Kriegner <dominik.kriegner@gmail.com>

import os.path
import unittest

import xrayutilities as xu

xu.config.VERBOSITY = 0
testfile = 'NISI.cif'
datadir = os.path.join(os.path.dirname(__file__), 'data')
fullfilename = os.path.join(datadir, testfile)


@unittest.skipIf(not os.path.isfile(fullfilename),
                 "additional test data needed (http://xrayutilities.sf.io)")
class TestIO_pdCIF(unittest.TestCase):
    dshape = (4495, )
    # dmax =  # needs update of pdCIF to handle value + error bar
    # dmin =
    motorname = '_pd_meas_time_of_flight'
    motmin = 1000.0
    motmax = 8190.4
    # dtpos =
    # tpos =

    @classmethod
    def setUpClass(cls):
        cls.dfile = xu.io.pdCIF(fullfilename)

    def test_datashape(self):
        self.assertEqual(self.dshape, self.dfile.data.shape)

    def test_datavalues(self):
        motor = self.dfile.data[self.motorname]
        # counter = self.dfile.data
        self.assertAlmostEqual(self.motmax, motor.max(), places=6)
        self.assertAlmostEqual(self.motmin, motor.min(), places=6)
        # self.assertAlmostEqual(self.dmax, counter.max(), places=6)
        # self.assertAlmostEqual(self.dmin, counter.min(), places=6)
        # self.assertAlmostEqual(counter[self.tpos[0], self.tpos[1]],
        #                        self.dtpos)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2012-2020 Dominik Kriegner <dominik.kriegner@gmail.com>
"""
module with a function wrapper to scipy.optimize.leastsq
for fitting of a 2D function to a peak or a 1D Gauss fit with
the odr package
"""

import time
import warnings

import numpy
import scipy.optimize as optimize
from scipy.odr import odrpack as odr

from .. import config, utilities
from ..exception import InputError
from .functions import (Gauss1d, Gauss1d_der_p, Gauss1d_der_x, Lorentz1d,
                        Lorentz1d_der_p, Lorentz1d_der_x, PseudoVoigt1d,
                        PseudoVoigt1d_der_p, PseudoVoigt1d_der_x,
                        PseudoVoigt1dasym, PseudoVoigt1dasym2)
from .misc import center_of_mass, fwhm_exp


def linregress(x, y):
    """
    fast linregress to avoid usage of scipy.stats which is slow!
    NaN values in y are ignored by this function.

    Parameters
    ----------
    x, y :  array-like
        data coordinates and values

    Returns
    -------
    p :     tuple
        parameters of the linear fit (slope, offset)
    rsq:    float
        R^2 value

    Examples
    --------
    >>> (k, d), R2 = xu.math.linregress(x, y)
    """
    mask = numpy.logical_not(numpy.isnan(y))
    lx, ly = (x[mask], y[mask])
    if numpy.all(numpy.isclose(lx-lx[0], numpy.zeros_like(lx))):
        return (0, numpy.mean(ly)), 0
    else:
        p = numpy.polyfit(lx, ly, 1)

        # calculation of r-squared
        f = numpy.polyval(p, lx)
        fbar = numpy.sum(ly) / len(ly)
        ssreg = numpy.sum((f-fbar)**2)
        sstot = numpy.sum((ly - fbar)**2)
        rsq = ssreg / sstot

        return p, rsq


def peak_fit(xdata, ydata, iparams=[], peaktype='Gauss', maxit=300,
             background='constant', plot=False, func_out=False, debug=False):
    """
    fit function using odr-pack wrapper in scipy similar to
    https://github.com/tiagopereira/python_tips/wiki/Scipy%3A-curve-fitting
    for Gauss, Lorentz or Pseudovoigt-functions

    Parameters
    ----------
    xdata :     array_like
        x-coordinates of the data to be fitted
    ydata :     array_like
        y-coordinates of the data which should be fit

    iparams :   list, optional
        initial paramters, determined automatically if not specified
    peaktype :  {'Gauss', 'Lorentz', 'PseudoVoigt', 'PseudoVoigtAsym', 'PseudoVoigtAsym2'}, optional
        type of peak to fit
    maxit :     int, optional
        maximal iteration number of the fit
    background : {'constant', 'linear'}, optional
        type of background function
    plot :      bool or str, optional
        flag to ask for a plot to visually judge the fit. If plot is a string
        it will be used as figure name, which makes reusing the figures easier.
    func_out :  bool, optional
        returns the fitted function, which takes the independent variables as
        only argument (f(x))

    Returns
    -------
    params :    list
        the parameters as defined in function `Gauss1d/Lorentz1d/PseudoVoigt1d/
        PseudoVoigt1dasym`. In the case of linear background one more parameter
        is included!
    sd_params : list
        For every parameter the corresponding errors are returned.
    itlim :     bool
        flag to tell if the iteration limit was reached, should be False
    fitfunc :   function, optional
        the function used in the fit can be returned (see func_out).
    """
    if plot:
        plot, plt = utilities.import_matplotlib_pyplot('XU.math.peak_fit')

    gfunc, gfunc_dx, gfunc_dp = _getfit_func(peaktype, background)

    # determine initial parameters
    _check_iparams(iparams, peaktype, background)
    if not any(iparams):
        iparams = _guess_iparams(xdata, ydata, peaktype, background)
    if config.VERBOSITY >= config.DEBUG:
        print("XU.math.peak_fit: iparams: %s" % str(tuple(iparams)))

    # set up odr fitting
    peak = odr.Model(gfunc, fjacd=gfunc_dx, fjacb=gfunc_dp)

    sy = numpy.sqrt(ydata)
    sy[sy == 0] = 1
    mydata = odr.RealData(xdata, ydata, sy=sy)

    myodr = odr.ODR(mydata, peak, beta0=iparams, maxit=maxit)
    myodr.set_job(fit_type=2)  # use least-square fit

    fit = myodr.run()
    if config.VERBOSITY >= config.DEBUG:
        print('XU.math.peak_fit:')
        fit.pprint()  # prints final message from odrpack

    fparam = fit.beta
    etaidx = []
    if peaktype in ('PseudoVoigt', 'PseudoVoigtAsym'):
        if background == 'linear':
            etaidx = [-2, ]
        else:
            etaidx = [-1, ]
    elif peaktype == 'PseudoVoigtAsym2':
        etaidx = [5, 6]
    for e in etaidx:
        fparam[e] = 0 if fparam[e] < 0 else fparam[e]
        fparam[e] = 1 if fparam[e] > 1 else fparam[e]

    itlim = False
    if fit.stopreason[0] == 'Iteration limit reached':
        itlim = True
        if config.VERBOSITY >= config.INFO_LOW:
            print("XU.math.peak_fit: Iteration limit reached, "
                  "do not trust the result!")

    if plot:
        if isinstance(plot, str):
            plt.figure(plot)
        else:
            plt.figure('XU:peak_fit')
        plt.plot(xdata, ydata, 'ok', label='data', mew=2)
        if debug:
            plt.plot(xdata, gfunc(iparams, xdata), '-', color='0.5',
                     label='estimate')
        plt.plot(xdata, gfunc(fparam, xdata), '-r',
                 label='%s-fit' % peaktype)
        plt.legend()

    if func_out:
        return fparam, fit.sd_beta, itlim, lambda x: gfunc(fparam, x)
    else:
        return fparam, fit.sd_beta, itlim


def _getfit_func(peaktype, background):
    """
    internal function to prepare the model functions and derivatives for the
    peak_fit function.

    Parameters
    ----------
    peaktype :  {'Gauss', 'Lorentz', 'PseudoVoigt', 'PseudoVoigtAsym', 'PseudoVoigtAsym2'}
        type of peak function
    background : {'constant', 'linear'}
        type of background function

    Returns
    -------
    f, f_dx, f_dp : functions
        fit function, function of derivative regarding `x`, and functions of
        derivatives regarding the parameters
    """
    gfunc_dx = None
    gfunc_dp = None
    if peaktype == 'Gauss':
        f = Gauss1d
        fdx = Gauss1d_der_x
        fdp = Gauss1d_der_p
    elif peaktype == 'Lorentz':
        f = Lorentz1d
        fdx = Lorentz1d_der_x
        fdp = Lorentz1d_der_p
    elif peaktype == 'PseudoVoigt':
        f = PseudoVoigt1d
        fdx = PseudoVoigt1d_der_x
        fdp = PseudoVoigt1d_der_p
    elif peaktype == 'PseudoVoigtAsym':
        if background == 'linear':
            def gfunc(param, x):
                return PseudoVoigt1dasym(x, *param) + x * param[-1]
        else:
            def gfunc(param, x):
                return PseudoVoigt1dasym(x, *param)
    elif peaktype == 'PseudoVoigtAsym2':
        if background == 'linear':
            def gfunc(param, x):
                return PseudoVoigt1dasym2(x, *param) + x * param[-1]
        else:
            def gfunc(param, x):
                return PseudoVoigt1dasym2(x, *param)
    else:
        raise InputError("keyword argument peaktype takes invalid value!")

    if peaktype in ('Gauss', 'Lorentz', 'PseudoVoigt'):
        if background == 'linear':
            def gfunc(param, x):
                return f(x, *param) + x * param[-1]

            def gfunc_dx(param, x):
                return fdx(x, *param) + param[-1]

            def gfunc_dp(param, x):
                return numpy.vstack((fdp(x, *param), x))
        else:
            def gfunc(param, x):
                return f(x, *param)

            def gfunc_dx(param, x):
                return fdx(x, *param)

            def gfunc_dp(param, x):
                return fdp(x, *param)
    return gfunc, gfunc_dx, gfunc_dp


def _check_iparams(iparams, peaktype, background):
    """
    internal function to check if the length of the supplied initial
    parameters is correct given the other settings of the peak_fit function.
    An InputError is raised in case of wrong shape or value.

    Parameters
    ----------
    iparams :   list
        initial paramters for the fit
    peaktype :  {'Gauss', 'Lorentz', 'PseudoVoigt', 'PseudoVoigtAsym', 'PseudoVoigtAsym2'}
        type of peak to fit
    background : {'constant', 'linear'}
        type of background

    """
    if not any(iparams):
        return
    else:
        ptypes = {('Gauss', 'constant'): 4, ('Lorentz', 'constant'): 4,
                  ('Gauss', 'linear'): 5, ('Lorentz', 'linear'): 5,
                  ('PseudoVoigt', 'constant'): 5, ('PseudoVoigt', 'linear'): 6,
                  ('PseudoVoigtAsym', 'constant'): 6,
                  ('PseudoVoigtAsym', 'linear'): 7,
                  ('PseudoVoigtAsym2', 'constant'): 7,
                  ('PseudoVoigtAsym2', 'linear'): 8}
        if not all(numpy.isreal(iparams)):
            raise InputError("XU.math.peak_fit: all initial parameters need to"
                             "be real!")
        elif (peaktype, background) in ptypes:
            nparams = ptypes[(peaktype, background)]
            if len(iparams) != nparams:
                raise InputError("XU.math.peak_fit: %d initial parameters "
                                 "are needed for %s-peak with %s background."
                                 % (nparams, peaktype, background))
        else:
            raise InputError("XU.math.peak_fit: invalid peak (%s) or "
                             "background (%s)" % (peaktype, background))


def _guess_iparams(xdata, ydata, peaktype, background):
    """
    internal function to automatically esitmate peak parameters from the data,
    considering also the background type.

    Parameters
    ----------
    xdata :     array-like
        x-coordinates of the data to be fitted
    ydata :     array-like
        y-coordinates of the data which should be fit
    peaktype :  {'Gauss', 'Lorentz', 'PseudoVoigt', 'PseudoVoigtAsym', 'PseudoVoigtAsym2'}
        type of peak to fit
    background : {'constant', 'linear'}
        type of background, either

    Returns
    -------
     list of initial parameters estimated from the data
    """
    ld = numpy.empty(len(ydata))
    # estimate peak position
    ipos, ld, back, slope = center_of_mass(xdata, ydata, background,
                                           full_output=True)
    maxpos = xdata[numpy.argmax(ld)]
    avx = numpy.average(xdata)
    if numpy.abs(ipos - avx) < numpy.abs(maxpos-avx):
        ipos = maxpos  # use the estimate which is further from the center

    # estimate peak width
    sigma1 = numpy.sqrt(numpy.sum(numpy.abs((xdata - ipos) ** 2 * ld)) /
                        numpy.abs(numpy.sum(ld)))
    sigma2 = fwhm_exp(xdata, ld)/(2 * numpy.sqrt(2 * numpy.log(2)))
    sigma = sigma1 if sigma1 < sigma2 else sigma2

    # build initial parameters
    iparams = [ipos, sigma, numpy.max(ld), back]
    if peaktype in ['Lorentz', 'PseudoVoigt']:
        iparams[1] *= 2 * numpy.sqrt(2 * numpy.log(2))
    if peaktype in ['PseudoVoigtAsym', 'PseudoVoigtAsym2']:
        iparams.insert(1, iparams[1])
    if peaktype in ['PseudoVoigt', 'PseudoVoigtAsym']:
        # set ETA parameter to be between Gauss and Lorentz shape
        iparams.append(0.5)
    if peaktype == 'PseudoVoigtAsym2':
        iparams.append(0.5)
        iparams.append(0.5)
    if background == 'linear':
        iparams.append(slope)
    return iparams


def gauss_fit(xdata, ydata, iparams=[], maxit=300):
    """
    Gauss fit function using odr-pack wrapper in scipy similar to
    https://github.com/tiagopereira/python_tips/wiki/Scipy%3A-curve-fitting

    Parameters
    ----------
    xdata :     array-like
        x-coordinates of the data to be fitted
    ydata :     array-like
        y-coordinates of the data which should be fit
    iparams:    list, optional
        initial paramters for the fit, determined automatically if not given
    maxit :     int, optional
        maximal iteration number of the fit

    Returns
    -------
    params :    list
        the parameters as defined in function ``Gauss1d(x, *param)``
    sd_params : list
        For every parameter the corresponding errors are returned.
    itlim :     bool
        flag to tell if the iteration limit was reached, should be False
    """
    return peak_fit(xdata, ydata, iparams=iparams,
                    peaktype='Gauss', maxit=maxit)


def fit_peak2d(x, y, data, start, drange, fit_function, maxfev=2000):
    """
    fit a two dimensional function to a two dimensional data set
    e.g. a reciprocal space map

    Parameters
    ----------
    x, y :      array-like
        data coordinates (do NOT need to be regularly spaced)
    data :      array-like
        data set used for fitting (e.g. intensity at the data coords)
    start :     list
        set of starting parameters for the fit used as first parameter of
        function fit_function
    drange :    list
        limits for the data ranges used in the fitting algorithm, e.g. it is
        clever to use only a small region around the peak which should be
        fitted: [xmin, xmax, ymin, ymax]
    fit_function : callable
        function which should be fitted, must be of form accept the parameters
        ``fit_function (x, y, *params) -> ndarray``

    Returns
    -------
    fitparam :  list
        fitted parameters
    cov :       array-like
        covariance matrix
    """
    s = time.time()
    if config.VERBOSITY >= config.INFO_ALL:
        print("XU.math.fit: Fitting started... ", end='')

    start = numpy.array(start)
    lx = x.flatten()
    ly = y.flatten()
    mask = (lx > drange[0]) * (lx < drange[1]) * \
        (ly > drange[2]) * (ly < drange[3])
    ly = ly[mask]
    lx = lx[mask]
    ldata = data.flatten()[mask]

    def errfunc(p, x, z, data):
        return fit_function(x, z, *p) - data

    p, cov, infodict, errmsg, success = optimize.leastsq(
        errfunc, start, args=(lx, ly, ldata), full_output=1, maxfev=maxfev)

    s = time.time() - s
    if config.VERBOSITY >= config.INFO_ALL:
        print("finished in %8.2f sec, (data length used %d)" % (s, ldata.size))
        print("XU.math.fit: %s" % errmsg)

    # calculate correct variance covariance matrix
    if cov is not None:
        s_sq = (errfunc(p, lx, ly, ldata) ** 2).sum() / \
            (len(ldata) - len(start))
        pcov = cov * s_sq
    else:
        pcov = numpy.zeros((len(start), len(start)))

    if success not in [1, 2, 3, 4]:
        print("XU.math.fit: Could not obtain fit!")
    return p, pcov


def multPeakFit(x, data, peakpos, peakwidth, dranges=None,
                peaktype='Gaussian', returnerror=False):
    """
    function to fit multiple Gaussian/Lorentzian peaks with linear background
    to a set of data

    Parameters
    ----------
    x :     array-like
        x-coordinate of the data
    data :  array-like
        data array with same length as `x`
    peakpos : list
        initial parameters for the peak positions
    peakwidth : list
        initial values for the peak width
    dranges : list of tuples
        list of tuples with (min, max) value of the data ranges to use.  does
        not need to have the same number of entries as peakpos
    peaktype : {'Gaussian', 'Lorentzian'}
        type of peaks to be used
    returnerror : bool
        decides if the fit errors of pos, sigma, and amp are returned (default:
        False)

    Returns
    -------
    pos :       list
        peak positions derived by the fit
    sigma :     list
        peak width derived by the fit
    amp :       list
        amplitudes of the peaks derived by the fit
    background :    array-like
        background values at positions `x`
    if returnerror == True:
     sd_pos :   list
        standard error of peak positions as returned by scipy.odr.Output
     sd_sigma : list
        standard error of the peak width
     sd_amp :   list
        standard error of the peak amplitude
    """
    warnings.warn("deprecated function -> use the lmfit Python packge instead",
                  DeprecationWarning)
    if peaktype == 'Gaussian':
        pfunc = Gauss1d
        pfunc_derx = Gauss1d_der_x
    elif peaktype == 'Lorentzian':
        pfunc = Lorentz1d
        pfunc_derx = Lorentz1d_der_x
    else:
        raise ValueError('wrong value for parameter peaktype was given')

    def deriv_x(p, x):
        """
        function to calculate the derivative of the signal of multiple peaks
        and background w.r.t. the x-coordinate

        Parameters
        ----------
        p :     list
            parameters, for every peak there needs to be position, sigma,
            amplitude and at the end two values for the linear background
            function (b0, b1)
        x :     array-like
            x-coordinate
        """
        derx = numpy.zeros(x.size)

        # sum up peak functions contributions
        for i in range(len(p) // 3):
            ldx = pfunc_derx(x, p[3 * i], p[3 * i + 1], p[3 * i + 2], 0)
            derx += ldx

        # background contribution
        k = p[-2]
        b = numpy.ones(x.size) * k

        return derx + b

    def deriv_p(p, x):
        """
        function to calculate the derivative of the signal of multiple peaks
        and background w.r.t. the parameters

        Parameters
        ----------
        p :     list
            parameters, for every peak there needs to be position, sigma,
            amplitude and at the end two values for the linear background
            function (b0, b1)
        x :     array-like
            x-coordinate

        returns derivative w.r.t. all the parameters with shape (len(p),x.size)
        """

        derp = numpy.empty(0)
        # peak functions contributions
        for i in range(len(p) // 3):
            lp = (p[3 * i], p[3 * i + 1], p[3 * i + 2], 0)
            if peaktype == 'Gaussian':
                derp = numpy.append(derp, -2 * (lp[0] - x) * pfunc(x, *lp))
                derp = numpy.append(
                    derp, (lp[0] - x) ** 2 / (2 * lp[1] ** 3) * pfunc(x, *lp))
                derp = numpy.append(derp, pfunc(x, *lp) / lp[2])
            else:  # Lorentzian
                derp = numpy.append(derp, 4 * (x - lp[0]) * lp[2] / lp[1] /
                                    (1 + (2 * (x - lp[0]) / lp[1]) ** 2) ** 2)
                derp = numpy.append(derp, 4 * (lp[0] - x) * lp[2] /
                                    lp[1] ** 2 / (1 + (2 * (x - lp[0]) /
                                                       lp[1]) ** 2) ** 2)
                derp = numpy.append(derp,
                                    1 / (1 + (2 * (x - p[0]) / p[1]) ** 2))

        # background contributions
        derp = numpy.append(derp, x)
        derp = numpy.append(derp, numpy.ones(x.size))

        # reshape output
        derp.shape = (len(p),) + x.shape
        return derp

    def fsignal(p, x):
        """
        function to calculate the signal of multiple peaks and background

        Parameters
        ----------
        p :     list
            list of parameters, for every peak there needs to be position,
            sigma, amplitude and at the end two values for the linear
            background function (k, d)
        x :     array-like
            x-coordinate
        """
        f = numpy.zeros(x.size)

        # sum up peak functions
        for i in range(len(p) // 3):
            lf = pfunc(x, p[3 * i], p[3 * i + 1], p[3 * i + 2], 0)
            f += lf

        # background
        k = p[-2]
        d = p[-1]
        b = numpy.polyval((k, d), x)

        return f + b

    ##########################
    # create local data set (extract data ranges)
    if dranges:
        mask = numpy.array([False] * x.size)
        for i in range(len(dranges)):
            lrange = dranges[i]
            lmask = numpy.logical_and(x > lrange[0], x < lrange[1])
            mask = numpy.logical_or(mask, lmask)
        lx = x[mask]
        ldata = data[mask]
    else:
        lx = x
        ldata = data

    # create initial parameter list
    p = []

    # background
    # exclude +/-2 peakwidth around the peaks
    bmask = numpy.ones_like(lx, dtype=bool)
    for pp, pw in zip(peakpos, peakwidth):
        bmask = numpy.logical_and(bmask, numpy.logical_or(lx < (pp-2*pw),
                                                          lx > (pp+2*pw)))
    if numpy.any(bmask):
        k, d = numpy.polyfit(lx[bmask], ldata[bmask], 1)
    else:
        if(config.VERBOSITY >= config.DEBUG):
            print("XU.math.multPeakFit: no data outside peak regions!")
        k, d = (0, ldata.min())

    # peak parameters
    for i in range(len(peakpos)):
        amp = ldata[(lx - peakpos[i]) >= 0][0] - \
            numpy.polyval((k, d), lx)[(lx - peakpos[i]) >= 0][0]
        p += [peakpos[i], peakwidth[i], amp]

    # background parameters
    p += [k, d]

    if(config.VERBOSITY >= config.DEBUG):
        print("XU.math.multPeakFit: intial parameters")
        print(p)

    ##########################
    # fit with odrpack
    model = odr.Model(fsignal, fjacd=deriv_x, fjacb=deriv_p)
    odata = odr.RealData(lx, ldata)
    my_odr = odr.ODR(odata, model, beta0=p)
    # fit type 2 for least squares
    my_odr.set_job(fit_type=2)
    fit = my_odr.run()

    if(config.VERBOSITY >= config.DEBUG):
        print("XU.math.multPeakFit: fitted parameters")
        print(fit.beta)
    try:
        if fit.stopreason[0] not in ['Sum of squares convergence']:
            print("XU.math.multPeakFit: fit NOT converged (%s)"
                  % fit.stopreason[0])
            return Nono, None, None, None
    except IndexError:
        print("XU.math.multPeakFit: fit most probably NOT converged (%s)"
              % str(fit.stopreason))
        return None, None, None, None
    # prepare return values
    fpos = fit.beta[:-2:3]
    fwidth = numpy.abs(fit.beta[1:-2:3])
    famp = fit.beta[2::3]
    background = numpy.polyval((fit.beta[-2], fit.beta[-1]), x)
    if returnerror:
        sd_pos = fit.sd_beta[:-2:3]
        sd_width = fit.sd_beta[1:-2:3]
        sd_amp = fit.sd_beta[2::3]
        return fpos, fwidth, famp, background, sd_pos, sd_width, sd_amp
    return fpos, fwidth, famp, background


def multPeakPlot(x, fpos, fwidth, famp, background, dranges=None,
                 peaktype='Gaussian', fig="xu_plot", ax=None, fact=1.):
    """
    function to plot multiple Gaussian/Lorentz peaks with background values
    given by an array

    Parameters
    ----------
    x :         array-like
        x-coordinate of the data
    fpos :      list
        positions of the peaks
    fwidth :    list
        width of the peaks
    famp :      list
        amplitudes of the peaks
    background :  array-like
        background values, same shape as `x`
    dranges :   list of tuples
        list of (min, max) values of the data ranges to use.  does not need to
        have the same number of entries as fpos
    peaktype : {'Gaussian', 'Lorentzian'}
        type of peaks to be used
    fig :       int, str, or None
        matplotlib figure number or name
    ax :        matplotlib.Axes
        matplotlib axes as alternative to the figure name
    fact :      float
        factor to use as multiplicator in the plot
    """
    warnings.warn("deprecated function -> use the lmfit Python packge instead",
                  DeprecationWarning)
    success, plt = utilities.import_matplotlib_pyplot('XU.math.multPeakPlot')
    if not success:
        return

    if fig:
        plt.figure(fig)
    if ax:
        plt.sca(ax)
    # plot single peaks
    if dranges:
        mask = numpy.array([False] * x.size)
        for i in range(len(dranges)):
            lrange = dranges[i]
            lmask = numpy.logical_and(x > lrange[0], x < lrange[1])
            mask = numpy.logical_or(mask, lmask)
        lx = x[mask]
        lb = background[mask]
    else:
        lx = x
        lb = background

    f = numpy.zeros(lx.size)
    for i in range(len(fpos)):
        if peaktype == 'Gaussian':
            lf = Gauss1d(lx, fpos[i], fwidth[i], famp[i], 0)
        elif peaktype == 'Lorentzian':
            lf = Lorentz1d(lx, fpos[i], fwidth[i], famp[i], 0)
        else:
            raise ValueError('wrong value for parameter peaktype was given')
        f += lf
        plt.plot(lx, (lf + lb) * fact, ':k')

    # plot summed signal
    plt.plot(lx, (f + lb) * fact, '-r', lw=1.5)



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2018 Dominik Kriegner <dominik.kriegner@gmail.com>

import xrayutilities as xu
from matplotlib.pylab import *

mpl.rcParams['font.size'] = 16.0

en = 'CuKa1'  # eV
resol = 0.0004  # resolution in q

sub = xu.simpack.Layer(xu.materials.Si, inf)
lay = xu.simpack.Layer(xu.materials.SiGe(0.6), 145.87, relaxation=0.0)
# pseudomorphic stack -> adjusts lattice parameters!
pls = xu.simpack.PseudomorphicStack001('pseudo', sub, lay)

H, K, L = (0, 0, 4)
qz = linspace(4.0, 7.0, 3000)

# calculate incidence angle for dynamical diffraction models
qx = sqrt(pls[0].material.Q(H, K, L)[0]**2 + pls[0].material.Q(H, K, L)[1]**2)
ai = xu.simpack.coplanar_alphai(qx, qz, en)
resolai = abs(xu.simpack.coplanar_alphai(qx, mean(qz) + resol, en) -
              xu.simpack.coplanar_alphai(qx, mean(qz), en))

# kinematic multibeam diffraction model
mk = xu.simpack.KinematicalMultiBeamModel(pls, energy=en,
                                          surface_hkl=(0, 0, 1),
                                          resolution_width=resol)
Imult = mk.simulate(qz, hkl=(H, K, L), refraction=True)

# general 2-beam theory based dynamical diffraction model
md = xu.simpack.DynamicalModel(pls, energy=en, resolution_width=resolai)
Idyn = md.simulate(ai, hkl=(H, K, L))

# plot of calculated intensities
figure('XU-simpack SiGe 004')
clf()
semilogy(qz, Imult, label='multibeam')
semilogy(xu.simpack.get_qz(qx, ai, en), Idyn, label='full dynamical')
legend(fontsize='small')
xlim(qz.min(), qz.max())
xlabel(r'Qz ($1/\mathrm{\AA}$)')
ylabel('Intensity (arb. u.)')
tight_layout()
show()

H, K, L = (2, 2, 4)
qz = linspace(4.0, 7.0, 3000)

# calculate incidence angle for dynamical diffraction models
qx = sqrt(pls[0].material.Q(H, K, L)[0]**2 + pls[0].material.Q(H, K, L)[1]**2)
ai = xu.simpack.coplanar_alphai(qx, qz, en)
resolai = abs(xu.simpack.coplanar_alphai(qx, mean(qz) + resol, en) -
              xu.simpack.coplanar_alphai(qx, mean(qz), en))

# kinematic multibeam diffraction model
mk = xu.simpack.KinematicalMultiBeamModel(pls, energy=en,
                                          surface_hkl=(0, 0, 1),
                                          resolution_width=resol)
Imult = mk.simulate(qz, hkl=(H, K, L), refraction=True)

# general 2-beam theory based dynamical diffraction model
md = xu.simpack.DynamicalModel(pls, energy=en, resolution_width=resolai)
Idyn = md.simulate(ai, hkl=(H, K, L))

# plot of calculated intensities
figure('XU-simpack SiGe 224')
clf()
semilogy(qz, Imult, label='multibeam')
semilogy(xu.simpack.get_qz(qx, ai, en), Idyn, label='full dynamical')
legend(fontsize='small')
xlim(qz.min(), qz.max())
xlabel(r'Qz ($1/\mathrm{\AA}$)')
ylabel('Intensity (arb. u.)')
tight_layout()
show()

H, K, L = (1, 1, 5)
qz = linspace(4.0, 7.0, 3000)

# calculate incidence angle for dynamical diffraction models
qx = sqrt(pls[0].material.Q(H, K, L)[0]**2 + pls[0].material.Q(H, K, L)[1]**2)
ai = xu.simpack.coplanar_alphai(qx, qz, en)
resolai = abs(xu.simpack.coplanar_alphai(qx, mean(qz) + resol, en) -
              xu.simpack.coplanar_alphai(qx, mean(qz), en))

# kinematic multibeam diffraction model
mk = xu.simpack.KinematicalMultiBeamModel(pls, energy=en,
                                          surface_hkl=(0, 0, 1),
                                          resolution_width=resol)
Imult = mk.simulate(qz, hkl=(H, K, L), refraction=True)

# general 2-beam theory based dynamical diffraction model
md = xu.simpack.DynamicalModel(pls, energy=en, resolution_width=resolai)
Idyn = md.simulate(ai, hkl=(H, K, L))

# plot of calculated intensities
figure('XU-simpack SiGe 115')
clf()
semilogy(qz, Imult, label='multibeam')
semilogy(xu.simpack.get_qz(qx, ai, en), Idyn, label='full dynamical')
legend(fontsize='small')
xlim(qz.min(), qz.max())
xlabel(r'Qz ($1/\mathrm{\AA}$)')
ylabel('Intensity (arb. u.)')
tight_layout()
show()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016 Dominik Kriegner <dominik.kriegner@gmail.com>

import unittest

import numpy
import xrayutilities as xu

digits = 9


class TestPeakFit(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.sl, cls.back = numpy.random.rand(2) * 0.1
        cls.width = numpy.random.rand() * 0.05 + 0.025
        cls.amp = numpy.random.rand() + 0.7
        cls.pos = numpy.random.rand() * 0.3 + 0.3
        cls.x = numpy.linspace(-0.1, 0.9, 1000)
        p = [cls.pos, cls.width, cls.amp, cls.back]
        cls.fg = xu.math.Gauss1d(cls.x, *p)
        cls.fl = xu.math.Lorentz1d(cls.x, *p)
        cls.eta = numpy.random.rand()
        p += [cls.eta]
        cls.fv = xu.math.PseudoVoigt1d(cls.x, *p)

    def test_gaussfit(self):
        params, sd_params, itlim = xu.math.gauss_fit(self.x, self.fg)
        self.assertAlmostEqual(params[0], self.pos, places=digits)
        self.assertAlmostEqual(abs(params[1]), self.width, places=digits)
        self.assertAlmostEqual(params[2], self.amp, places=digits)
        self.assertAlmostEqual(params[3], self.back, places=digits)

    def test_gaussfit_linear(self):
        f = self.fg + self.sl * self.x
        params, sd_params, itlim, ffunc = xu.math.peak_fit(
            self.x, f, peaktype='Gauss', background='linear', func_out=True)
        self.assertAlmostEqual(params[0], self.pos, places=digits)
        self.assertAlmostEqual(abs(params[1]), self.width, places=digits)
        self.assertAlmostEqual(params[2], self.amp, places=digits)
        self.assertAlmostEqual(params[3], self.back, places=digits)
        self.assertAlmostEqual(params[4], self.sl, places=digits)

    def test_lorentzfit(self):
        params, sd_params, itlim = xu.math.peak_fit(
            self.x, self.fl, peaktype='Lorentz', background='constant')
        self.assertAlmostEqual(params[0], self.pos, places=digits)
        self.assertAlmostEqual(abs(params[1]), self.width, places=digits)
        self.assertAlmostEqual(params[2], self.amp, places=digits)
        self.assertAlmostEqual(params[3], self.back, places=digits)

    def test_lorentzfit_linear(self):
        iparam = numpy.asarray([self.pos, self.width, self.amp, self.back,
                                self.sl])
        iparam += (numpy.random.rand(5) - 0.5) * 0.05
        f = self.fl + self.sl * self.x
        params, sd_params, itlim = xu.math.peak_fit(
            self.x, f, peaktype='Lorentz', background='linear', iparams=iparam)
        self.assertAlmostEqual(params[0], self.pos, places=digits)
        self.assertAlmostEqual(abs(params[1]), self.width, places=digits)
        self.assertAlmostEqual(params[2], self.amp, places=digits)
        self.assertAlmostEqual(params[3], self.back, places=digits)
        self.assertAlmostEqual(params[4], self.sl, places=digits)

    def test_pvoigtfit(self):
        params, sd_params, itlim = xu.math.peak_fit(
            self.x, self.fv, peaktype='PseudoVoigt', background='constant')
        self.assertAlmostEqual(params[0], self.pos, places=digits)
        self.assertAlmostEqual(abs(params[1]), self.width, places=digits)
        self.assertAlmostEqual(params[2], self.amp, places=digits)
        self.assertAlmostEqual(params[3], self.back, places=digits)
        self.assertAlmostEqual(params[4], self.eta, places=digits)

    def test_pvoigtfit_linear(self):
        f = self.fv + self.sl * self.x
        params, sd_params, itlim = xu.math.peak_fit(
            self.x, f, peaktype='PseudoVoigt', background='linear')
        self.assertAlmostEqual(params[0], self.pos, places=digits)
        self.assertAlmostEqual(abs(params[1]), self.width, places=digits)
        self.assertAlmostEqual(params[2], self.amp, places=digits)
        self.assertAlmostEqual(params[3], self.back, places=digits)
        self.assertAlmostEqual(params[4], self.eta, places=digits)
        self.assertAlmostEqual(params[5], self.sl, places=digits)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2018-2020 Dominik Kriegner <dominik.kriegner@gmail.com>

import os
import tempfile
import unittest

import xrayutilities as xu

xu.config.VERBOSITY = 0


class Test_CIF_export(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.materials = []
        for name, obj in xu.materials.predefined_materials.__dict__.items():
            if isinstance(obj, xu.materials.Crystal):
                cls.materials.append(obj)

    def test_export(self):
        for m in self.materials:
            with tempfile.NamedTemporaryFile(mode='w', delete=False) as fid:
                filename = fid.name
            m.toCIF(filename)
            c = xu.materials.Crystal.fromCIF(filename)
            self.assertEqual(m, c, msg=f"Crystal {m.name} different after "
                             "toCIF/fromCIF conversion")
            os.remove(filename)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2015 Dominik Kriegner <dominik.kriegner@gmail.com>

import os.path
import unittest

import xrayutilities as xu

xu.config.VERBOSITY = 0  # make no output during test
testfile = 'speclog.log.gz'
datadir = os.path.join(os.path.dirname(__file__), 'data')
fullfilename = os.path.join(datadir, testfile)


@unittest.skipIf(not os.path.isfile(fullfilename),
                 "additional test data needed (http://xrayutilities.sf.io)")
class TestIO_SPECLog(unittest.TestCase):
    prompt = 'PSIC'
    line_cnt = 2046542
    testcmd = '84.PSIC>  mvr mu -.075; ct'
    testcmdline = 25

    def setUp(self):
        self.logfile = xu.io.SPECLog(testfile, self.prompt, path=datadir)

    def test_linenumber(self):
        self.assertEqual(self.line_cnt, self.logfile.line_counter)
        self.assertEqual(self.testcmd, str(self.logfile[self.testcmdline]))


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2018 Dominik Kriegner <dominik.kriegner@gmail.com>

# ALSO LOOK AT THE FILE xrayutilities_id01_functions.py

"""
example script for determining ESRF/ID01 detector parameters using a specfile
from the second half of 2017
"""

import os
import re

import xrayutilities as xu

import xrayutilities_id01_functions as id01

s = xu.io.SPECFile(specfile)  # insert specfile name here
specscan = s.scan3
en = id01.getmono_energy(specscan)
# template for the CCD file names
filetmp = id01.getmpx4_filetmp(specscan)

images = []
ang1 = specscan.data['nu']
ang2 = specscan.data['del']

# read images and angular positions from the data file
for imgnr in imagenrs:
    filename = filetmp % imgnr
    edf = xu.io.EDFFile(filename)
    images.append(edf.data)

# call the fit for the detector parameters
# detector arm rotations and primary beam direction need to be given
param, eps = xu.analysis.sample_align.area_detector_calib(
    ang1, ang2, images, ['z-', 'y-'], 'x+',
    start=(55e-6, 55e-6, 0.5, 45, 0, -0.7, 0),
    fix=(True, True, False, False, False, False, True),
    wl=xu.en2lam(en))



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2020 Dominik Kriegner <dominik.kriegner@gmail.com>

import itertools
import re
import unittest

import numpy
import xrayutilities as xu
from xrayutilities.materials import SymOp


class TestMaterialsSymOp(unittest.TestCase):
    @classmethod
    def setUp(cls):
        cls.x, cls.y, cls.z = numpy.random.rand(3)
        a, b, c = numpy.random.rand(3) * 2 + 4
        al, be, gam = numpy.random.rand(3) * 60 + 60
        pdict = {'a': a, 'b': b, 'c': c, 'alpha': al, 'beta': be, 'gamma': gam}

        cls.lats = []
        cls.gps = []
        for sg in xu.materials.spacegrouplattice.wp.keys():
            # determine parameters for this space group
            sgnr = int(sg.split(':')[0])
            csys, nargs = xu.materials.spacegrouplattice.sgrp_sym[sgnr]
            params = xu.materials.spacegrouplattice.sgrp_params[csys][0]
            p = [eval(par, pdict) for par in params]
            # generate test lattice
            cls.lats.append(xu.materials.SGLattice(sg, *p))
            # get general Wyckoff position of space group
            wp = xu.materials.spacegrouplattice.wp
            gplabel = sorted(wp[sg], key=lambda s: int(s[:-1]))[-1]
            cls.gps.append(wp[sg][gplabel][1])

    def test_SymOp_fromxyz(self):
        """
        tests that the creation of the symmetry operations from the xyz string
        of the Wyckoff positions is reproducible/reversable
        """
        for lat, gp in zip(self.lats, self.gps):
            symopsxyz = map(lambda s: '({})'.format(s.xyz()), lat.symops)
            self.assertCountEqual(symopsxyz, gp)

    def test_equivalent_hkl(self):
        hkl = numpy.random.randint(-11, 12, 3)
        for lat in self.lats:
            ehkl = numpy.unique(numpy.einsum('...ij,j', lat._hklsym, hkl),
                                axis=0)
            ehkl = set(tuple(e) for e in ehkl)
            self.assertEqual(lat.equivalent_hkls(hkl), ehkl)

    def test_iscentrosymmetric(self):
        centrosym = list(itertools.chain([2], range(10, 16), range(47, 75),
                                         range(83, 89), range(123, 143),
                                         [147, 148], range(162, 168),
                                         [175, 176], range(191, 195),
                                         range(200, 207), range(221, 231)))
        for lat in self.lats:
            if lat.space_group_nr in centrosym:
                self.assertTrue(lat.iscentrosymmetric)
            else:
                self.assertFalse(lat.iscentrosymmetric)

    def test_Wyckoff_consistency(self):
        """
        tests that all Wyckoff positions are consistent with the symmetry
        operations
        """
        reint = re.compile('[0-9]+')
        pardict = {'x': self.x, 'y': self.y, 'z': self.z}
        wp = xu.materials.spacegrouplattice.wp
        for lat, gp in zip(self.lats, self.gps):
            # check every Wyckoff position
            for wpkey in wp[lat.space_group]:
                uniquepos = []
                poscount = int(reint.match(wpkey).group())
                thispositions = list(
                    map(lambda p: SymOp.foldback(eval(p, pardict)),
                        wp[lat.space_group][wpkey][1]))
                pos0 = SymOp.foldback(thispositions[0])
                genpos = [s.apply(pos0) for s in lat.symops]
                uniquepos = [genpos[0], ]
                for p in genpos:
                    considered = False
                    for u in uniquepos:
                        if numpy.allclose(p, u):
                            considered = True
                    if not considered:
                        uniquepos.append(p)
                comparepos = [(t, g) for t in thispositions for g in uniquepos
                              if numpy.allclose(t, g)]
                # check that number of Wyckoff position entries are correct
                self.assertEqual(poscount, len(thispositions))
                # check that an equal number of unique positions is produced
                self.assertEqual(poscount, len(uniquepos))
                # check that unique positions are equal to Wyckoff positions
                self.assertEqual(poscount, len(comparepos))


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2015 Dominik Kriegner <dominik.kriegner@gmail.com>

import matplotlib as mpl
import matplotlib.pyplot as plt
import xrayutilities as xu
from matplotlib.colors import LogNorm

# mpl >= 3.3.0
# mpl.rcParams["pcolor.shading"] = 'auto'

# global setting for the experiment
sample = "rsm"  # sample name used also as file name for the data file

# substrate material used for Bragg peak calculation to correct for
# experimental offsets
Si = xu.materials.Si

hxrd = xu.HXRD(Si.Q(1, 1, 0), Si.Q(0, 0, 1))

#################################
# Si/SiGe (004) reciprocal space map
omalign = 34.3051  # experimental aligned values
ttalign = 69.1446
# nominal values of the substrate peak
[omnominal, _, _, ttnominal] = hxrd.Q2Ang(Si.Q(0, 0, 4))

# read the data from the xrdml files
om, tt, psd = xu.io.getxrdml_map(sample + '_%d.xrdml.bz2', 1,
                                 path='data')

# convert angular coordinates to reciprocal space + correct for offsets
[qx, qy, qz] = hxrd.Ang2Q(om, tt, delta=[omalign - omnominal,
                                         ttalign - ttnominal])

#############################################
# three different visualization possibilities
# plot the intensity as contour plot
plt.figure(figsize=(12, 6))
ax = []
for i in range(1, 4):
    ax.append(plt.subplot(1, 3, i))

MIN = 1
MAX = 3e5
plt.sca(ax[0])
plt.title('Gridder2D')
# data on a regular grid of 200x800 points
gridder = xu.Gridder2D(200, 300)
gridder(qy, qz, psd)
cf = plt.pcolormesh(gridder.xaxis, gridder.yaxis, gridder.data.T,
                    norm=LogNorm(MIN, MAX))

plt.sca(ax[1])
plt.title('FuzzyGridder2D')
# data on a regular grid with FuzzyGridding
gridder = xu.FuzzyGridder2D(200, 300)
gridder(qy, qz, psd, width=(0.0008, 0.0003))
cf = plt.pcolormesh(gridder.xaxis, gridder.yaxis, gridder.data.T,
                    norm=LogNorm(MIN, MAX))

plt.sca(ax[2])
plt.title('pcolormesh')
# using pcolor-variants
npixel = 255
qy.shape = (qy.size//npixel, npixel)
qz.shape = qy.shape
psd.shape = qy.shape
plt.pcolormesh(qy, qz, psd, norm=LogNorm(MIN, MAX))

for a in ax:
    plt.sca(a)
    plt.xlabel(r'$Q_{[110]}$ ($\mathrm{\AA}^{-1}$)')
    plt.ylabel(r'$Q_{[001]}$ ($\mathrm{\AA}^{-1}$)')
    plt.xlim(-0.13, 0.13)
    plt.ylim(4.538, 4.654)
    plt.xticks((-0.1, 0.0, 0.1))
plt.tight_layout()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2015 Dominik Kriegner <dominik.kriegner@gmail.com>

import unittest

import numpy
import xrayutilities as xu


class TestFuzzyGridder3D(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.n = 10
        cls.min = 1
        cls.max = 10
        cls.axis = numpy.linspace(cls.min, cls.max, cls.n)
        cls.x = int(numpy.random.rand()*10) + 1
        cls.y = int(numpy.random.rand()*10) + 1
        cls.z = int(numpy.random.rand()*10) + 1
        cls.data = numpy.random.rand()
        cls.gridder = xu.FuzzyGridder3D(cls.n, cls.n, cls.n)
        cls.gridder.dataRange(cls.min, cls.max,
                              cls.min, cls.max,
                              cls.min, cls.max)
        cls.gridder(cls.x, cls.y, cls.z, cls.data, width=2)

    def test_gridder3d_xaxis(self):
        # test length of xaxis
        self.assertEqual(len(self.gridder.xaxis), self.n)
        # test values of xaxis
        for i in range(self.n):
            self.assertAlmostEqual(self.gridder.xaxis[i], self.axis[i],
                                   places=12)

    def test_gridder3d_yaxis(self):
        # test length of yaxis
        self.assertEqual(len(self.gridder.yaxis), self.n)
        # test end values of yaxis
        for i in range(self.n):
            self.assertAlmostEqual(self.gridder.yaxis[i], self.axis[i],
                                   places=12)

    def test_gridder3d_zaxis(self):
        # test length of zaxis
        self.assertEqual(len(self.gridder.zaxis), self.n)
        # test end values of yaxis
        for i in range(self.n):
            self.assertAlmostEqual(self.gridder.zaxis[i], self.axis[i],
                                   places=12)

    def test_fuzzygridder3d_data(self):
        # test shape of data
        self.assertEqual(self.gridder.data.shape[0], self.n)
        self.assertEqual(self.gridder.data.shape[1], self.n)
        self.assertEqual(self.gridder.data.shape[2], self.n)
        # test values of data
        vg = numpy.zeros((self.gridder.data.shape))
        norm = numpy.copy(vg)
        ix, iy, iz = self.x - 1, self.y - 1, self.z - 1
        for i in range(ix-1, ix+2):
            for j in range(iy-1, iy+2):
                for k in range(iz-1, iz+2):
                    idx1 = i
                    idx2 = j
                    idx3 = k
                    n = 1/8.
                    if abs(i - ix) > 0:
                        n /= 2.
                    if abs(j - iy) > 0:
                        n /= 2.
                    if abs(k - iz) > 0:
                        n /= 2.
                    if i < 0:
                        idx1 = 0
                    elif i >= self.n:
                        idx1 = -1
                    if j < 0:
                        idx2 = 0
                    elif j >= self.n:
                        idx2 = -1
                    if k < 0:
                        idx3 = 0
                    elif k >= self.n:
                        idx3 = -1
                    norm[idx1, idx2, idx3] += n
                    vg[idx1, idx2, idx3] = self.data

        for i in range(self.n):
            for j in range(self.n):
                for k in range(self.n):
                    self.assertAlmostEqual(self.gridder.data[i, j, k],
                                           vg[i, j, k],
                                           places=12)
                    self.assertAlmostEqual(self.gridder._gnorm[i, j, k],
                                           norm[i, j, k],
                                           places=12)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016 Dominik Kriegner <dominik.kriegner@gmail.com>

import unittest

import numpy
import xrayutilities as xu

digits = 10


class TestSolveQuartic(unittest.TestCase):
    z = [(-2.1120821587966274-2.945550430762482j),
         (-2.1120821587966274+2.945550430762482j),
         0.6655848599174902,
         2.0585794576757652]

    def test_solver(self):
        roots = xu.math.solve_quartic(2, 3, 6, -60, 36)
        roots = numpy.sort(roots)

        for i, root in enumerate(roots):
            self.assertAlmostEqual(root, self.z[i], places=digits)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2019 Dominik Kriegner <dominik.kriegner@gmail.com>

import math
import os.path
import unittest

import xrayutilities as xu

xu.config.VERBOSITY = 0  # make no output during test
testfile = 'sardana2.8.spec'
datadir = os.path.join(os.path.dirname(__file__), 'data')
fullfilename = os.path.join(datadir, testfile)


@unittest.skipIf(not os.path.isfile(fullfilename),
                 "additional test data needed (http://xrayutilities.sf.io)")
class TestIO_SPEC_Sardana(unittest.TestCase):
    scannrs = [4, 6, 8, 10]
    s4_init_motor_pos_len = 6
    s8_init_motor_pos_len = 2
    s8_init_mopo_mot12 = 20.0
    s8_dshape = (29,)
    s8_ncol = 9
    s8_dtmax = 10.7484171391
    s8_dtmin = 1.35413002968
    s8_mot15max = -4.4
    s8_mot15min = -10.0
    s10_pos = 119
    s10_dpos = 119
    countername = 'Pt_No'

    @classmethod
    def setUpClass(cls):
        cls.specfile = xu.io.SPECFile(testfile, path=datadir)
        cls.scans = dict()
        for nr in cls.scannrs:
            cls.scans[nr] = getattr(cls.specfile, 'scan%d' % nr)
            cls.scans[nr].ReadData()

    def test_init_mopo(self):
        self.assertEqual(self.s4_init_motor_pos_len,
                         len(self.scans[4].init_motor_pos))
        self.assertEqual(self.s8_init_motor_pos_len,
                         len(self.scans[8].init_motor_pos))
        self.assertAlmostEqual(
            self.s8_init_mopo_mot12,
            self.scans[8].init_motor_pos['INIT_MOPO_mot12'], places=6)
        with self.assertRaises(KeyError):
            # test not existing initial motor position
            self.scans[8].init_motor_pos['INIT_MOPO_offset03']

    def test_datashape_aborted(self):
        self.assertEqual(self.s8_dshape, self.scans[8].data.shape)
        self.assertEqual(self.s8_ncol, len(self.scans[8].colnames))

    def test_datavalues(self):
        self.assertTrue(math.isnan(self.scans[6].data['ct13'][29]))
        s8d = self.scans[8].data
        self.assertAlmostEqual(self.s8_dtmax, s8d['dt'].max(), places=6)
        self.assertAlmostEqual(self.s8_dtmin, s8d['dt'].min(), places=6)
        self.assertAlmostEqual(self.s8_mot15max, s8d['mot15'].max(), places=6)
        self.assertAlmostEqual(self.s8_mot15min, s8d['mot15'].min(), places=6)
        s10d = self.scans[10].data
        self.assertAlmostEqual(self.s10_dpos,
                               s10d[self.countername][self.s10_pos], places=6)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2014 Dominik Kriegner <dominik.kriegner@gmail.com>

"""
This example shows the use of the Q2AngFit function to calculate experimental
angles which can not be calculated by the analytic functions coded in the
HXRD, NonCOP and GID experimental classes, which use a fixed geometry.

Here an arbitrary goniometer together with some restrictions can be defined and
experimental angles can be calculated for this geometry
"""

import time

import numpy
import xrayutilities as xu

numpy.set_printoptions(precision=4, threshold=100, suppress=True)

energy = 15000

###########################
# definition of goniometer
###########################
# 3S+2D goniometer (simplified ID01 goniometer, sample mu,eta,phi detector
# nu,del
qconv = xu.experiment.QConversion(['z+', 'y-', 'z-'], ['z+', 'y-'], [1, 0, 0])
# convention for coordinate system: x downstream; z upwards; y to the "outside"
# (righthanded)
# QConversion will set up the goniometer geometry. So the first argument
# describes the sample rotations, the second the detector rotations and the
# third the primary beam direction.
# For this consider the following coordinate system (at least this is what i
# use at ID01, feel free to use your conventions):
# x: downstream (direction of primary beam)
# y: out of the ring
# z: upwards
# these three axis form a right handed coordinate system.
# The outer most sample rotation (so the one mounted on the floor) is one which
# turns righthanded (+) around the z-direction -> z+ (at the moment this
# rotation is called 'mu' in the spec-session)
# The second sample rotation ('eta') is lefthanded (-) around y -> y-

# define experimental geometry with respect to the crystalline directions
# of the substrate
hxrd = xu.HXRD((1, 0, 0), (0, 0, 1), en=energy, qconv=qconv)

# tell bounds of angles / (min,max) pair for all motors
# mu,eta,phi detector nu,del
bounds = (0, (0, 90), 0, (-1, 90), (0, 90))


#############################
# call angle fit function
#############################

ang = None
tbegin = time.time()
print("time  error (code)      qvec             angles")
for i in range(100):

    qvec = numpy.array((0, 0, i * 0.01))
    t0 = time.time()
    ang, qerror, errcode = xu.Q2AngFit(qvec, hxrd, bounds, startvalues=ang)
    t1 = time.time()

    print("%.4f: %.3g (%d) %s %s" % (t1 - t0, qerror, errcode,
                                     str(qvec), str(ang)))

tend = time.time()
print("Total time needed: %.2fsec" % (tend - tbegin))



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016-2020 Dominik Kriegner <dominik.kriegner@gmail.com>
"""
simulation subpackage of xrayutilities.

This package provides possibilities to simulate X-ray diffraction and
reflectivity curves of thin film samples. It could be extended for more
general use in future if there is demand for that.

In addition it provides a fitting routine for reflectivity data which is based
on lmfit.
"""

from .darwin_theory import (DarwinModel, DarwinModelAlGaAs001,
                            DarwinModelAlloy, DarwinModelGaInAs001,
                            DarwinModelSiGe001, GradedBuffer)
from .fit import FitModel
from .helpers import coplanar_alphai, get_qz
from .models import (DiffuseReflectivityModel, DynamicalModel,
                     DynamicalReflectivityModel, KinematicalModel,
                     KinematicalMultiBeamModel, LayerModel, Model,
                     SimpleDynamicalCoplanarModel, SpecularReflectivityModel,
                     effectiveDensitySlicing)
from .mosaicity import mosaic_analytic
from .powder import FP_profile, PowderDiffraction
from .powdermodel import PowderModel, Rietveld_error_metrics, plot_powder
from .smaterials import (CrystalStack, GradedLayerStack, Layer, LayerStack,
                         MaterialList, Powder, PowderList,
                         PseudomorphicStack001, PseudomorphicStack111,
                         SMaterial)



"""
Example script to show how to use xrayutilities to read and plot
reciprocal space map scans from a spec file created at the ESRF/ID10B

for details about the measurement see:
    D Kriegner et al. Nanotechnology 22 425704 (2011)
    http://dx.doi.org/10.1088/0957-4484/22/42/425704
"""

import os

import matplotlib.pyplot as plt
import numpy
import xrayutilities as xu

# global setting for the experiment
sample = "test"  # sample name used also as file name for the data file
energy = 8042.5  # x-ray energy in eV
center_ch = 715.9  # center channel of the linear detector
chpdeg = 345.28  # channels per degree of the linear detector
roi = [100, 1340]  # region of interest of the detector
nchannel = 1500  # number of channels of the detector

# intensity normalizer function responsible for count time and absorber
# correction
normalizer_detcorr = xu.IntensityNormalizer(
    "MCA",
    mon="Monitor",
    time="Seconds",
    absfun=lambda d: d["detcorr"] / d["psd2"].astype(numpy.float))

# substrate material used for Bragg peak calculation to correct for
# experimental offsets
InP = xu.materials.InP

# initialize experimental class to specify the reference directions of your
# crystal
# 11-2: inplane reference
# 111: surface normal
hxrd = xu.HXRD(InP.Q(1, 1, -2), InP.Q(1, 1, 1), en=energy)

# configure linear detector
# detector direction + parameters need to be given
# mounted along z direction, which corresponds to twotheta
hxrd.Ang2Q.init_linear('z-', center_ch, nchannel, chpdeg=chpdeg, roi=roi)

# read spec file and save to HDF5-file
# since reading is much faster from HDF5 once the data are transformed
h5file = os.path.join("data", sample + ".h5")
try:
    s  # try if spec file object already exist ("run -i" in ipython)
except NameError:
    s = xu.io.SPECFile(sample + ".spec", path="data")
else:
    s.Update()
s.Save2HDF5(h5file)

#################################
# InP (333) reciprocal space map
omalign = 43.0529  # experimental aligned values
ttalign = 86.0733
[omnominal, _, _, ttnominal] = hxrd.Q2Ang(
    InP.Q(3, 3, 3))  # nominal values of the substrate peak

# read the data from the HDF5 file
# scan number:36, names of motors in spec file: omega= sample rocking, gamma =
# twotheta
[om, tt], MAP = xu.io.geth5_scan(h5file, 36, 'omega', 'gamma')
# normalize the intensity values (absorber and count time corrections)
psdraw = normalizer_detcorr(MAP)
# remove unusable detector channels/regions (no averaging of detector channels)
psd = xu.blockAveragePSD(psdraw, 1, roi=roi)

# convert angular coordinates to reciprocal space + correct for offsets
[qx, qy, qz] = hxrd.Ang2Q.linear(
    om, tt,
    delta=[omalign - omnominal, ttalign - ttnominal])

# calculate data on a regular grid of 200x201 points
gridder = xu.Gridder2D(200, 201)
gridder(qy, qz, psd)
# maplog function limits the shown dynamic range to 8 orders of magnitude
# from the maxium
INT = xu.maplog(gridder.data.T, 8., 0)

# plot the intensity as contour plot using matplotlib
plt.figure()
cf = plt.contourf(gridder.xaxis, gridder.yaxis, INT, 100, extend='min')
plt.xlabel(r'$Q_{[11\bar2]}$ ($\mathrm{\AA}^{-1}$)')
plt.ylabel(r'$Q_{[\bar1\bar1\bar1]}$ ($\mathrm{\AA}^{-1}$)')
cb = plt.colorbar(cf)
cb.set_label(r"$\log($Int$)$ (cps)")



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016 Dominik Kriegner <dominik.kriegner@gmail.com>

import os.path
import unittest

import xrayutilities as xu

xu.config.VERBOSITY = 0  # make no output during test
testfile = 'fastscan.spec.gz'
datadir = os.path.join(os.path.dirname(__file__), 'data')
fullfilename = os.path.join(datadir, testfile)


@unittest.skipIf(not os.path.isfile(fullfilename),
                 "additional test data needed (http://xrayutilities.sf.io)")
class TestIO_FastScan(unittest.TestCase):
    dshape = (62500,)
    dshape_retrace = (61992,)
    dmax = 154795.0
    dmin = 3797.0
    xmax = 80.038498
    xmin = 40.283199
    ymin = 138.61099
    ymax = 180.02901
    cname = 'mpx4int'
    timer = 'timer'
    timermax = 1285.15

    @classmethod
    def setUp(cls):
        cls.fs = xu.io.FastScan(testfile, 3, path=datadir)

    def test_grid2D(self):
        N = 200
        self.fs.retrace_clean()
        g2d = self.fs.grid2D(N, N+1, gridrange=((self.xmin, self.xmax),
                                                (self.ymin, self.ymax)))
        self.assertEqual(g2d.data.shape, (N, N+1))
        self.assertAlmostEqual(g2d.xaxis.min(), self.xmin, places=4)
        self.assertAlmostEqual(g2d.xaxis.max(), self.xmax, places=4)
        self.assertAlmostEqual(g2d.yaxis.min(), self.ymin, places=4)
        self.assertAlmostEqual(g2d.yaxis.max(), self.ymax, places=4)

    def test_datashape(self):
        self.assertEqual(self.fs.data.shape, self.dshape)
        self.assertEqual(self.fs.xvalues.shape, self.dshape)
        self.assertEqual(self.fs.yvalues.shape, self.dshape)
        self.fs.retrace_clean()
        self.assertEqual(self.fs.data.shape, self.dshape_retrace)
        self.assertEqual(self.fs.xvalues.shape, self.dshape_retrace)
        self.assertEqual(self.fs.yvalues.shape, self.dshape_retrace)

    def test_datavalues(self):
        self.assertAlmostEqual(self.fs.motorposition(self.timer).max(),
                               self.timermax, places=2)
        self.assertAlmostEqual(self.fs.xvalues.min(), self.xmin, places=5)
        self.assertAlmostEqual(self.fs.xvalues.max(), self.xmax, places=5)
        self.assertAlmostEqual(self.fs.yvalues.min(), self.ymin, places=5)
        self.assertAlmostEqual(self.fs.yvalues.max(), self.ymax, places=5)
        self.assertAlmostEqual(self.fs.data[self.cname].min(), self.dmin,
                               places=5)
        self.assertAlmostEqual(self.fs.data[self.cname].max(), self.dmax,
                               places=5)


if __name__ == '__main__':
    unittest.main()
